! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                            Domain                            !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Module for a 3D computational domain.
!!
!! Naming conventions for spatial 3D position, size, and bounds of a block, domain, etc.:
!!  - *position* - $pos(3): start indices in global coordinates, e.g. (3, 1, 5)
!!  - *extent*   - $ext(3): number of grid cells (local coordinates), e.g. (5, 5, 5) for a block containing 5**3=125 cells.
!!    This corresponds to the extent of a dimension in Fortran speech.
!!  - *bounds*   - $bnd(3,2): combination of $pos and $ext, contains start ($pos) and end indices in global coordinates, e.g.
!!    (3, 1, 5), (7, 5, 9).
!!    This corresponds to the bounds of a dimension in Fortran speech.
!!
!! Following conversions are valid:
!!  - $pos = $bnd(:,1) 
!!  - $ext = $bnd(:,2) - $bnd(:,1) + 1
!!
!! A character prepended to a variable name for $pos, $ext, or $bnd means:
!!  - $d - domain (e.g. $dbnd contains the start and end indices of a domain)
!!  - $b - block  (e.g. $bpos contains the start indices of a block)
!!
!!
!! Preprocessor options:
!!  - $FD4_ADAPT_BGB: allocate boundary ghost blocks in $fd4_balance_readjust (only if necessary - block%empty),
!!    default is to allocate boundary ghost blocks immediately after block creation
!!  - $FD4_DOMAIN_NO_POOL: disable block pools
!!
!! @author Matthias Lieber
module fd4_domain_mod

  use fd4_block_mod
  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_ghostlist_mod
  use fd4_mpi_mod
  use fd4_stat_mod
  use fd4_metadata_mod
  use rbtree_fd4_block_mod
  use rbtree_int8_mod
  use stack_mod
  use stack3_mod
  use section_mod
  use timing_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH

  
  !! shifting lengths for $fd4_domain_c2i and $fd4_domain_i2c
  integer, parameter, public :: c2i_shft3 = 20, c2i_shft2 = 10
  
  !! number of history records of load balancing to gain estimates for auto load balancing mode
  integer, parameter, public :: FD4_LB_HISTORY = 4
   
  ! public types
  public :: rbtree_fd4_block, fd4_vartab, fd4_varinfo, fd4_mpi
  public :: fd4_domain, fd4_domain_coord, fd4_block, fd4_ghostlist
  public :: fd4_domain_balance, fd4_domain_blpool
  public :: stack, section, rbtree_int8, fd4_block_link, stack3
  public :: fd4_domain_lb_params

  ! public methods
  public :: fd4_domain_create
  public :: fd4_domain_delete
  public :: fd4_domain_add_block
  public :: fd4_domain_delete_block
  public :: fd4_domain_clear
  public :: fd4_domain_max_bext
  public :: fd4_domain_add_ghost
  public :: fd4_domain_delete_ghost
  public :: fd4_domain_dump_stats
  public :: fd4_domain_init_metadata
  public :: fd4_domain_adapt_metadata

  !! Create new domain.
  !!
  interface fd4_domain_create
    module procedure fd4_domain_create_fixed, fd4_domain_create_specific
  end interface fd4_domain_create


  !! Type to store data related to dynamic load balancing
  !!
  !! *TODO*: put other domain stuff here, e.g. $empties, $sfclevel
  type fd4_domain_balance
    real(r4k) :: defaultweight = 1.0_r4k        !! default weight of new created blocks (per grid cell!)
    real(r4k), pointer :: blkweights(:,:,:) => NULL() !! weights of all blocks
    real(r8k), pointer :: blkweights_ps(:) => NULL()  !! prefix sum of all block weights for $fd4_part_sfc
    real(r4k) :: localweight                    !! sum of local block weights, updated in $fd4_part_sfc
    integer(i_k), pointer :: hidx(:) => NULL()  !! sfc index of global allocated blocks
    integer(i_k), pointer :: rnk(:)  => NULL()  !! ranking of global allocated blocks by sfc index
    type(stack3) :: allblocks                   !! coordinates of all required blocks in global domain
    integer(i_k) :: last_sfctype = -1           !! last used type of SFC
    integer(i_k) :: sfc_pstart = -1             !! start index of local partition on SFC
    integer(i_k) :: map2d(2)                    !! mapping of 3D to 2D for 2D SFC
    ! stuff for auto balance mode
    !  timing history
    real(r4k) :: lbtimes(FD4_LB_HISTORY)        !! timings of last $FD4_LB_HISTORY calls of load balancing
    integer(i_k) :: curlbtime                   !! index of current entry in $lbtimes
    logical :: globalize_lbtime                 !! indicate whether current lbtime needs to be globalized (max)
    !  balance history
    real(r4k) :: lbbalance(FD4_LB_HISTORY)      !! balance achieved by last $FD4_LB_HISTORY calls of load balancing
    integer(i_k) :: curlbbalance                !! index of current entry in $lbbalance
    logical :: part_changed                     !! indicate whether fd4_balance_readjust tried to balance the load last time
    !  balance statistics
    integer(i_k) :: stat_check_us               !! index of balance checking time in the statistics object
    integer(i_k) :: stat_part_us                !! index of partition calculation time in the statistics object
    integer(i_k) :: stat_mig_us                 !! index of migration time in the statistics object
    integer(i_k) :: stat_migblock               !! index of migrated blocks in the statistics object
    integer(i_k) :: stat_migbytes               !! index of migrated bytes in the statistics object
    integer(i_k) :: stat_binsteps               !! index of maximum binary search steps for SFC methods
  end type fd4_domain_balance
  
  !! Type to store default partitioning method parameters
  !! (can be overriden when calling fd4_balance_readjust)
  type fd4_domain_lb_params
    logical       :: intialized = .false. !! tells fd4_balance wheater the parameter have been initialized
    integer(i_k)  :: method   !! partitioning method (SFC, ParMetis, RCB, etc.)
    real(r4k)     :: ubvec(1) !! ParMetis ubvec parameter
    real(r4k)     :: itr      !! ParMetis itr parameter
    integer(i_k)  :: subset   !! ParMetis: use only $subset processes for ParMetis calculation (<2...off)
    real(r4k)     :: wfactor  !! ParMetis: multiply this with block weights before converting to int
    integer(i_k)  :: sfctype  !! SFC: space-filling curve type
    logical       :: force    !! force to run ghost block allocation and deallocation    
    real(r4k)     :: lbtol    !! load balance tolerance, 0.0 (max. relaxed) ... 1.0 (strictest)
    logical       :: ech      !! if false, assume that empty info did not change in any block
    logical       :: auto     !! automatic decision whether load balancing is beneficial or not
    logical       :: weight   !! if false ignore block weights (i.e. assume each block has same weight)
    integer(i_k)  :: binsteps !! SFC: max. number of steps for binary search methods, default: 100
    real(r8k)     :: accuracy !! SFC: target ratio from best possible load balance for binary steps estimation, *must be <= 1.0*
  end type fd4_domain_lb_params


  !! Type for a spatial dimension, internally used by $fd4_domain
  type fd4_domain_coord
    integer(i_k), pointer :: bl(:) => null()          !! start indices of blocks
  end type fd4_domain_coord
  

  !! A pool of unused blocks, implemented as a linked list. Internally used by $fd4_domain.
  type fd4_domain_blpool
    type(fd4_block), pointer :: start => null()
    integer(i_k) :: len = 0
    integer(i_k) :: maxlen = 0
  end type fd4_domain_blpool
   

  !! Type for a 3D computational domain.
  !!
  !! @see fd4_vartab_mod
  !! @see rbtree_fd4_block_mod
  type fd4_domain
  
    type(rbtree_fd4_block) :: blocks                  !! data structure containing all blocks of the domain

    type(fd4_domain_coord) :: d(3)                    !! information about each of the 3 spatial dimensions
    
    integer(i_k) :: dbnd(3,2)                         !! start and end indices of the grid for x,y,z coordinates
    integer(i_k) :: blnum(3)                          !! number of blocks in x,y,z direction
    
    integer(i_k) :: mmbsz(3,2)                        !! min and max block size in each spatial dimension
    
    integer(i_k) :: nghosts(3)                        !! number of ghost cells for each spatial dimension
    logical      :: periodic(3)                       !! apply periodic boundary conditions?

    type(fd4_vartab), pointer :: vartab(:) => null()  !! variable table
    
    type(fd4_varinfo) :: varinfo                      !! global info about variable table

    type(fd4_ghostlist) :: glist(-1:0)                !! lists of all ghost blocks (boundary + communication ghosts)

    type(fd4_mpi) :: mpi                              !! MPI context

    integer, pointer :: blkmap(:,:,:,:) => null()     !! Who owns which block? Contains MPI ranks. With boundary.
    integer :: blkmap_now                             !! current block map (4th dimension of $blkmap)
    integer :: blkmap_last                            !! last block map (4th dimension of $blkmap)


    logical :: init = .false.                         !! true if domain is initialized
    integer(i_k) :: sfclevel                          !! level of the SFC (depends on blnum)
    
    integer(i8k) :: partition_id                      !! unique identifier of the current partition (counter)
    integer(i8k) :: partition_called                  !! counts up if a partitioning routine has been called

    integer(i4k), pointer :: empties(:,:,:)           !! block%sempty of all blocks
    
    integer(i_k) :: blockcount = 0                    !! global number of blocks
    type(section) :: blockunion                       !! spatial union of bounds of all local blocks
    
    type(fd4_domain_balance) :: balance               !! Stores data related to dynamic load balancing
    
    type(stack) :: markedblocks                       !! list of blocks (idx) marked as to be created (global)
    logical :: blocks_are_marked                      !! are there any blocks marked (global)
    
    type(fd4_domain_blpool), pointer :: pool(:,:,:)   !! block pools for each possible block size
    type(fd4_domain_blpool), pointer :: gpool(:,:,:)  !! ghost block pools for each possible ghost block size
    integer(i_k) :: gpmap(3)                          !! mapping of ghost block size to position in $gpool
    
    type(rbtree_int8) :: cpltypestore                 !! a binary tree to store MPI types for coupling
    integer(i4k) :: cplblkenc(14)                     !! encoding params of index for this tree
    
    type(fd4_domain_lb_params) :: lbparam             !! default load balancing method parameters
    
    type(fd4_stat) :: stat                            !! FD4 statistics and profiling
    logical :: stat_enabled = .true.                  !! switch statistics on/off
    integer(i_k) :: stat_bl_alloc                     !! index of allocated blocks in the statistics object
    integer(i_k) :: stat_bl_free                      !! index of deallocated blocks in the statistics object
    integer(i_k) :: stat_gb_alloc                     !! index of allocated ghost blocks in the statistics object
    integer(i_k) :: stat_gb_free                      !! index of deallocated ghost blocks in the statistics object
    
    integer(i_k) :: timer = TIMING_MOD_GETTIMEOFDAY   !! timer used for $fd4_iter_start_clock
    
    type(fd4_metadata) :: md                          !! meta data for coupling
    
  end type fd4_domain

contains

  include 'util/stack.incf'
  include 'util/stack3.incf'
  include 'util/math.incf'

  ! literally include this file here to have the routines in the generated documentation
  include 'framework/fd4_domain.incf'



! ============ Block functions ============ 


  !! Add block at given block coordinates.
  !!
  !! Allocates new block and block's static arrays. Adds the block to the 
  !! block data structure.
  !! Set up the neighbor pointers of this and neighbor blocks.
  !! The value for bpos(i) must be >=1 and <=domain%blnum(i).
  !!
  subroutine fd4_domain_add_block(domain, bpos, init, err, bl, opt_alloc, opt_dealloc)
  
    implicit none
    type(fd4_domain), intent(inout)     :: domain      !! the domain
    integer(i_k)    , intent(in)        :: bpos(3)     !! position of the block in the block decomposition
    logical         , intent(in)        :: init        !! if true, data fields are initialized with their vnull
    integer(i_k)    , intent(out)       :: err         !! error status: 0...ok, -1...block already present,  >0...error
    type(fd4_block),optional,pointer    :: bl          !! optionally return a pointer to the new block
    integer(i_k),optional, intent(inout):: opt_alloc   !! add here the number of real block allocations (no block in pool found)
    integer(i_k),optional, intent(inout):: opt_dealloc !! add here the number of real block deallocations
    
    
    type(fd4_block), pointer :: b
    type(fd4_domain_blpool), pointer :: pool
    integer(i_k) :: bext(3), hidx, i, dim, dir
    logical :: at_boundary, grabbed_from_pool
    integer(i_k) :: alloc, dealloc
    
    err = 1
    
    alloc = 0
    dealloc = 0
    
    FD4_L3_ASSERT(.not. domain%init,'domain empty, create one first')
    
    ! check block position in the global grid - bpos(3)
    do i=1,3
      FD4_L3_ASSERT(bpos(i)<1,'bpos < 1')
      FD4_L3_ASSERT(bpos(i)>domain%blnum(i),'bpos > domain%blnum')
    end do

    bext = fd4_domain_blext(domain, bpos)
      
    ! get index of block on SFC
    call fd4_domain_c2i(bpos, hidx)

    ! check if there's a block of the required size in the pool,
    ! so we don't need to do any memory allocation here
    pool => domain%pool(bext(1),bext(2),bext(3))
    grabbed_from_pool = pool%len > 0
    if( grabbed_from_pool ) then
      ! we can grab a block from the pool
      b => pool%start
      FD4_L3_ASSERT(.not.associated(b),'no block in pool')
      FD4_L3_ASSERT(ANY(b%ext/=bext),'wrong block extent')
      pool%start => b%neigh(1,1)%l
      b%neigh(1,1)%l => null()
      pool%len = pool%len - 1
    else
      ! we need to allocate a new block
      allocate(b, stat=err)
      FD4_L3_ASSERT(err/=0,'allocate block')
    end if

    ! add the new block to the data structure
    call rbtree_fd4_block_insert (domain%blocks, hidx, b, err)
    if(err/=0) then
      ! Block at this position already exists
      ! This is NO fatal error and may be intended by the caller!
      ! Delete the block and get out of here
      err = -1
      if( grabbed_from_pool ) then
        ! if we grabbed the block from the pool, we put it back
        pool%len = pool%len + 1
        b%neigh(1,1)%l => pool%start
        pool%start => b
      else
        deallocate(b)
      end if
      return
    end if

    ! set block's position in the block decomposition
    b%pos(:) = bpos(:)

    ! set block's global grid position
    do i=1,3
      b%bnd(i,:) = (/ domain%d(i)%bl(bpos(i)) , domain%d(i)%bl(bpos(i)+1) - 1 /)
    end do
    
    if( grabbed_from_pool ) then
      ! if we re-use a block, we should reset some fields
      call fd4_block_reset(b, domain%vartab, domain%varinfo, init)
      FD4_L5_INFO('grabbed block from pool at '//l2s(INT(hidx,i8k))//p2s(bpos))
    else
      ! allocate the data fields of the block
      call fd4_block_allocate(b, bext, domain%vartab, domain%varinfo, init, .true., .true., err)
      FD4_L3_ASSERT(err/=0,'fd4_block_allocate')
      FD4_L5_INFO('allocated block at '//l2s(INT(hidx,i8k))//p2s(bpos))
      alloc = alloc + 1
      if(domain%stat_enabled) then
        domain%stat%values(domain%stat_bl_alloc) = domain%stat%values(domain%stat_bl_alloc) + 1
      end if
    end if


    ! set up the neighbor pointers:    
    ! loop over (/ minus, plus /)
    do dir=1,2
      ! loop over 3 spatial dimensions
      do dim=1,3

        at_boundary = (bpos(dim)==1 .and. dir==1) .or. (bpos(dim)==domain%blnum(dim) .and. dir==2)

        if( at_boundary .and. .not. domain%periodic(dim) ) then
          ! boundary of the domain, but no periodic boundary conditions
#ifdef FD4_ADAPT_BGB
          ! boundary ghost block will be allocated later if really needed
          b%neigh(dim,dir)%l => null()
#else
          ! add boundary ghost block
          if(domain%nghosts(dim)>0) then
            call fd4_domain_add_ghost(domain, b, dir, dim, FD4_GHOST_BND, opt_alloc = alloc)
          end if
#endif
        else
          bext(:) = bpos(:)
          if(at_boundary) then
            ! apply periodic boundary conditions
            if(dir==1)  bext(dim) = domain%blnum(dim)
            if(dir==2)  bext(dim) = 1
          else
            bext(dim) = bext(dim) + fd4_incdec(dir)
          end if
          if( ALL( bext(1:3)==bpos(1:3) ) ) then
            ! oops! I'm my own neighbor. this happens, if domain%blnum(dim)==1
            b%neigh(dim,dir)%l => b
          else
            ! get index of this neighbor block
            call fd4_domain_c2i(bext, hidx)
            ! get pointer to neighbor block (or NULL if not present)
            b%neigh(dim,dir)%l => rbtree_fd4_block_get(domain%blocks, hidx)
            ! update neighbor's neighbor pointer
            if(associated(b%neigh(dim,dir)%l)) then
              if(associated(b%neigh(dim,dir)%l%neigh(dim,3-dir)%l)) then
                ! if this is a comm. ghost block, we don't need it anymore, let's remove it
                call fd4_domain_delete_ghost(domain, b%neigh(dim,dir)%l%neigh(dim,3-dir)%l, opt_dealloc = dealloc)
              end if
              b%neigh(dim,dir)%l%neigh(dim,3-dir)%l => b
            else
              ! has no neighbor
              ! communication ghost blocks are allocated in fd4_balance
              continue
            end if        
          end if  ! bext(:)==bpos(:)
        end if ! at_boundary .and. .not. domain%periodic(dim)

      end do  
    end do
    
    b%sempty = 0
    
    if(present(bl)) bl => b
    if(present(opt_alloc)) opt_alloc = opt_alloc + alloc
    if(present(opt_dealloc)) opt_dealloc = opt_dealloc + dealloc

    err = 0

  end subroutine fd4_domain_add_block
  

  !! Internal Routine to add a boundary or comm. ghost block near $b.
  !!
  !!  *Note*: Data fields of ghost blocks are not initialized with $vnull, since
  !!  is is expected that these data is read only after ghost communication.
  !!
  !!  Called from $fd4_domain_add_block and in $fd4_balance_readjust.
  !!  The new gost block $bgb is added to the ghostlist $glist
  !!  and a pointer from $bgb is set to $b:
  !!v bgb%neigh(2,1)%l => b
  !!
  !! Here is a list of all $block components which are abused by ghost blocks:
  !!>
  !!  bgb%neigh(1,1)%l  ...  previous in ghostlist
  !!  bgb%neigh(1,2)%l  ...  next in ghostlist
  !!  bgb%neigh(2,1)%l  ...  pointer to neighor within domain (ordinary block), "parent"
  !!  bgb%pos(1)        ...  boundary ghost block indicator, (FD4_GHOST_BND or FD4_GHOST_CMM)
  !!  bgb%pos(2)        ...  boundary direction (1 or 2)
  !!  bgb%pos(3)        ...  boundary dimension (1, 2 or 3)
  !!<  
  !!
  subroutine fd4_domain_add_ghost(domain, b, dir, dim, gtype, opt_alloc)
  
    implicit none
    type(fd4_domain)     , intent(inout) :: domain    !! the domain
    type(fd4_block)      , pointer       :: b         !! the block
    integer(i_k)         , intent(in)    :: dir       !! boundary ghost block direction
    integer(i_k)         , intent(in)    :: dim       !! boundary ghost block dimension
    integer(i_k)         , intent(in)    :: gtype     !! the type of ghost (FD4_GHOST_BND or FD4_GHOST_CMM)
    integer(i_k),optional, intent(inout) :: opt_alloc !! add here the number of real block allocations (no block in pool found)
    
    type(fd4_block) , pointer       :: bgb      !! the ghost block
    integer(i_k) :: bext(3), ext(3), err
    type(fd4_domain_blpool), pointer :: pool

    ! set ghost block's extents
    bext = b%ext
    bext(dim) = domain%nghosts(dim)    

    ! check if there's a block of the required size in the pool,
    ! so we don't need to do any memory allocation here
    ext = bext
    where(ext>domain%nghosts) ext = ext + domain%gpmap ! map block extent to position in ghost pool
    pool => domain%gpool(ext(1),ext(2),ext(3))
    if( pool%len > 0 ) then
      ! we can grab a block from the pool
      bgb => pool%start
      FD4_L3_ASSERT(.not.associated(bgb),'no block in pool')
      FD4_L3_ASSERT(ANY(bgb%ext/=bext),'wrong ghost block extent')
      pool%start => bgb%neigh(1,1)%l
      bgb%neigh(1,1)%l => null()
      pool%len = pool%len - 1
      ! reset some fields
      call fd4_block_reset(bgb, domain%vartab, domain%varinfo, .false.)
      FD4_L5_INFO('grabbed ghost block from pool at '//p2s(b%pos)//i2s(dim)//i2s(dir))
    else
      ! we need to allocate a new block
      allocate(bgb, stat=err)
      ! allocate the data fields
      call fd4_block_allocate(bgb, bext, domain%vartab, domain%varinfo, .false., .false., .false., err)
      FD4_L3_ASSERT(err/=0,'fd4_block_allocate')
      FD4_L5_INFO('allocated ghost block at '//p2s(b%pos)//i2s(dim)//i2s(dir))
      if(present(opt_alloc)) opt_alloc = opt_alloc + 1
      if(domain%stat_enabled) then
        domain%stat%values(domain%stat_gb_alloc) = domain%stat%values(domain%stat_gb_alloc) + 1
      end if
    end if

    ! we are using the pos array for some ghost block special information
    bgb%pos(1) = gtype     ! boundary/comm. ghost block indicator
    bgb%pos(2) = dir       ! ghost direction
    bgb%pos(3) = dim       ! ghost dimension

    bgb%bnd(:,:) = b%bnd(:,:)
    if(dir==1) then
      bgb%bnd(dim,2) = b%bnd(dim,1) - 1
      bgb%bnd(dim,1) = b%bnd(dim,1) - domain%nghosts(dim)
    else
      bgb%bnd(dim,1) = b%bnd(dim,2) + 1
      bgb%bnd(dim,2) = b%bnd(dim,2) + domain%nghosts(dim)
    end if
    
    b%neigh(dim,dir)%l => bgb
    bgb%neigh(2,1)%l => b

    ! add ghost block to the list
    call fd4_ghostlist_add(domain%glist(gtype), bgb)

  end subroutine fd4_domain_add_ghost



  !! Delete block at given index.
  !!
  !! Deletes all memory associated with the block, including
  !! both types of ghost blocks. Clears neighbor pointers in
  !! neighbor blocks.
  subroutine fd4_domain_delete_block(domain, hidx, err, opt_dealloc)
    implicit none
    type(fd4_domain)    , intent(inout) :: domain      !! the domain
    integer(i_k)        , intent(in)    :: hidx        !! the index of the block on the SFC
    integer(i_k)        , intent(out)   :: err         !! error status: 0...ok
    integer(i_k),optional,intent(inout) :: opt_dealloc !! add here the number of real block deallocations (no space left in pool)

    type(fd4_block), pointer :: b
    type(fd4_domain_blpool), pointer :: pool
    integer(i_k) :: dim, dir
    integer(i_k) :: gtype ! ghost type (bnd or cmm)
    integer(i_k) :: dealloc

    err = -1

    ! delete the block from the data structure
    b => rbtree_fd4_block_delete (domain%blocks, hidx)

    FD4_L3_ASSERT(.not. associated(b),'block not allocated')
    
    dealloc = 0

    ! clear the neighbor pointers in the neighboring blocks:
    ! loop over (/ minus, plus /)
    do dir=1,2
      ! loop over 3 spatial dimensions
      do dim=1,3
        ! nullify neighbor's neighbor pointer
        if(associated(b%neigh(dim,dir)%l)) then
          ! check if neighbor is a ghost block
          gtype = b%neigh(dim,dir)%l%pos(1)
          if(gtype <= 0) then
            ! delete ghost block
            call fd4_domain_delete_ghost(domain, b%neigh(dim,dir)%l, opt_dealloc = dealloc)
            b%neigh(dim,dir)%l => NULL()
          else
            b%neigh(dim,dir)%l%neigh(dim,3-dir)%l => null()
          end if
        end if        
      end do  
    end do
    
    ! push the block to the pool of unused blocks
    pool => domain%pool(b%ext(1),b%ext(2),b%ext(3))
    if( pool%len < pool%maxlen ) then
      b%neigh(1,1)%l => pool%start
      pool%start => b
      pool%len = pool%len + 1
      FD4_L5_INFO('pushed block to pool at  '//l2s(INT(hidx,i8k)))
    else
      ! no more space left in this pool, so deallocate the block
      call fd4_block_deallocate(b)
      FD4_L5_INFO('deleted block at  '//l2s(INT(hidx,i8k)))
      dealloc = dealloc + 1
      if(domain%stat_enabled) then
        domain%stat%values(domain%stat_bl_free) = domain%stat%values(domain%stat_bl_free) + 1
      end if
    end if
    
    if(present(opt_dealloc)) opt_dealloc = opt_dealloc + dealloc

    err = 0
       
  end subroutine fd4_domain_delete_block



  !! Internal routine to remove a boundary or comm. ghost block $gb.
  !!
  subroutine fd4_domain_delete_ghost(domain, gb, opt_dealloc)
    implicit none
    type(fd4_domain), intent(inout)     :: domain      !! the domain
    type(fd4_block) , pointer           :: gb          !! the ghost block
    integer(i_k),optional,intent(inout) :: opt_dealloc !! add here the number of real block deallocations (no space left in pool)
    
    integer(i_k) :: gtype, ext(3)
    type(fd4_domain_blpool), pointer :: pool
    
    FD4_L3_ASSERT(.not.associated(gb),'ghost block not assiciated')
    
    ! identify ghost type (comm or boundary)
    gtype = gb%pos(1) 
    FD4_L3_ASSERT(gtype>0.or.gtype<-1,'no ghost block')

    ! remove from ghostlist
    call fd4_ghostlist_remove(domain%glist(gtype), gb)
    
    ! push the block to the pool of unused blocks
    ext = gb%ext
    where(ext>domain%nghosts) ext = ext + domain%gpmap ! map block extent to position in ghost pool
    pool => domain%gpool(ext(1),ext(2),ext(3))
    if( pool%len < pool%maxlen ) then
      gb%neigh(1,1)%l => pool%start
      pool%start => gb
      pool%len = pool%len + 1
      FD4_L5_INFO('pushed ghost block to pool')
    else
      ! no more space left in this pool, so deallocate the block
      call fd4_block_deallocate(gb)
      FD4_L5_INFO('deleted ghost block')
      if(present(opt_dealloc)) opt_dealloc = opt_dealloc + 1
      if(domain%stat_enabled) then
        domain%stat%values(domain%stat_gb_free) = domain%stat%values(domain%stat_gb_free) + 1
      end if
    end if

  end subroutine fd4_domain_delete_ghost
  
  
  
  
  
  

! ============ Domain functions ============ 



  !! Get the extent of an array which is large enough to hold any single block in this domain.
  !! 
  !! This is not necessarily the extent of the largest block.
  !! The allocation status of the blocks does not influence the result of this routine.
  !!
  subroutine fd4_domain_max_bext(domain, bext, add_ghosts)
    implicit none
    type(fd4_domain) , intent(in)  :: domain     !! the domain
    integer(i_k)     , intent(out) :: bext(3)    !! block extents greater than or equal to any block's extents, 0 on error
    logical, optional, intent(in)  :: add_ghosts !! if present and true, adds ghost cells to the extent
    
    integer(i_k) :: i, j
    
    bext(:) = 0
    if(.not. domain%init) return
    
    do i=1,3
      do j=1,domain%blnum(i)
        bext(i) = MAX( bext(i), domain%d(i)%bl(j+1)-domain%d(i)%bl(j) )
      end do    
    end do
    
    if(present(add_ghosts)) then
      if(add_ghosts) bext(1:3) = bext(1:3) + 2 * domain%nghosts(1:3)
    end if
        
  end subroutine fd4_domain_max_bext
  


  !! Create domain with fixed block sizes.
  !!
  !! This function calculates the actual block bounds from the number of blocks
  !! in each spatial direction and the start and end indices of the grid.
  !!
  !! For example, the values
  !!> dbnd(:,1) = (/ 1,-3, 1/)    ! start indices of grid in x,y,z direction
  !!  dbnd(:,2) = (/ 9, 5, 3/)    ! end   indices of grid in x,y,z direction
  !!< blnum     = (/ 3, 3, 3/)    ! number of blocks in x,y,z direction
  !! result in the following block definitions used for fd4_domain_create_specific:
  !!> blx = (/ 1,  4,  7, 10/) 
  !!  bly = (/-3,  0,  3,  6/)
  !!< blz = (/ 1,  2,  3,  4/)
  !! @see fd4_vartab_mod
  subroutine fd4_domain_create_fixed(domain, blnum, dbnd, vars, nghosts, periodic, mpi_comm, err)

    implicit none
    
    type(fd4_domain), intent(inout) :: domain        !! the domain
    integer(i_k),     intent(in)    :: dbnd(3,2)     !! start and end indices of the grid for x,y,z coordinates
    integer(i_k)    , intent(in)    :: blnum(3)      !! number of blocks in x,y,z direction
    type(fd4_vartab), intent(in)    :: vars(:)       !! variable table
    integer(i_k),     intent(in)    :: nghosts(3)    !! number of required ghost cells for each spatial dimension
    logical,          intent(in)    :: periodic(3)   !! apply periodic boundary conditions?
    integer(i_k),     intent(in)    :: mpi_comm      !! MPI communicator
    integer(i_k),     intent(out)   :: err           !! error status: 0...ok
    
    integer(i_k) :: blx(blnum(1)+1), bly(blnum(2)+1), blz(blnum(3)+1)
    integer(i_k) :: i
    
    err = -1    
    
    do i=1,3
      if(dbnd(i,2)<dbnd(i,1)) then
        FD4_L1_ERROR('domain size <= 0')
      else if(dbnd(i,2)-dbnd(i,1)+1<blnum(i)) then
        FD4_L1_ERROR('domain size < number of blocks')
      else if(blnum(i)<1) then
        FD4_L1_ERROR('number of blocks < 1')
      end if
    end do
    
    blx(1)=dbnd(1,1)
    do i=1,blnum(1)
       blx(i+1)=(i*(dbnd(1,2)+1-blx(1)))/blnum(1)+blx(1)
    end do

    bly(1)=dbnd(2,1)
    do i=1,blnum(2)
       bly(i+1)=(i*(dbnd(2,2)+1-bly(1)))/blnum(2)+bly(1)
    end do

    blz(1)=dbnd(3,1)
    do i=1,blnum(3)
       blz(i+1)=(i*(dbnd(3,2)+1-blz(1)))/blnum(3)+blz(1)
    end do
    
    !print *,blx
    !print *,bly
    !print *,blz
    
    call fd4_domain_create_specific(domain, blx, bly, blz, vars, nghosts, periodic, mpi_comm, err)    

  end subroutine fd4_domain_create_fixed




  !! Create domain with specific block sizes.
  !!
  !!  The array $blx is of length $num_x_blocks + 1. The first $num_x_blocks
  !!  values define the start indices of the blocks in x direction. The last
  !!  value defines the start index of an additional imaginary block and thus
  !!  the size of the last real block. The same applies for $bly and $blz.
  !!  The bounds of the grid is derived from the block definitions.
  !!  For example
  !!> blx = (/ 1,  4,  7, 10/) 
  !!  bly = (/-3,  0,  3,  6/)
  !!< blz = (/ 1,  2,  3,  4/)
  !!  define a grid with x dimension from 1 to 9, y dimension from -3 to 5
  !!  and z dimension from 1 to 3. The grid is subdivided in 3 x 3 x 3 blocks
  !!  which have the size 3 x 3 x 1 grid cells.
  !!
  !! @see fd4_vartab_mod
  subroutine fd4_domain_create_specific(domain, blx, bly, blz, vars, nghosts, periodic, mpi_comm, err)
    
    implicit none
    
    type(fd4_domain), intent(inout) :: domain          !! the domain
    integer(i_k),     intent(in)    :: blx(:)          !! start indices of blocks in x direction
    integer(i_k),     intent(in)    :: bly(:)          !! start indices of blocks in y direction
    integer(i_k),     intent(in)    :: blz(:)          !! start indices of blocks in z direction
    type(fd4_vartab), intent(in)    :: vars(:)         !! variable table
    integer(i_k),     intent(in)    :: nghosts(3)      !! number of required ghost cells for each spatial dimension
    logical,          intent(in)    :: periodic(3)     !! apply periodic boundary conditions?
    integer(i_k),     intent(in)    :: mpi_comm        !! MPI communicator
    integer(i_k),     intent(out)   :: err             !! error status: 0...ok
    
    integer(i_k) :: i, bl, minext, ierr, dim
    character(len=3), parameter :: dimstr(0:3) = (/'  -', '  x', '  y', '  z' /)
    character(len=10) :: vthres
    type(fd4_vartab), pointer :: v
    integer(i_k) :: x, y, z, wsize, ext(3)
    integer(i_k), allocatable :: nbsz(:,:,:), ngbsz(:,:,:)
    integer(i_k) :: gbbnd(3,2), bbnd(3,2), nstacks, totstacksize
#ifndef FD4_DOMAIN_NO_POOL
    real :: nprocs13
#endif   

    err = -1    

    ! get global MPI rank for output line head
    call MPI_Comm_rank(MPI_COMM_WORLD,fd4_wrnk,ierr)
    call MPI_Comm_size(MPI_COMM_WORLD,wsize,ierr)
    if(                     wsize<10000   ) write(FD4_LINEHD,'(A,I4.4,A)') '[FD4:',fd4_wrnk,'] '
    if(  10000<=wsize .and. wsize<100000  ) write(FD4_LINEHD,'(A,I5.5,A)') '[FD:', fd4_wrnk,'] '
    if( 100000<=wsize .and. wsize<1000000 ) write(FD4_LINEHD,'(A,I6.6,A)') '[F:',  fd4_wrnk,'] '
    if(1000000<=wsize                     ) write(FD4_LINEHD,'(A,I8.8,A)') '[',    fd4_wrnk,'] '
    
    ! check if domain is empty (i.e. no allocated fields)
    FD4_L1_ASSERT(domain%init,'domain already in use')
    
    ! check if number of blocks is > 1
    FD4_L1_ASSERT(size(blx)<2,'no x-blocks defined')
    FD4_L1_ASSERT(size(bly)<2,'no y-blocks defined')
    FD4_L1_ASSERT(size(blz)<2,'no z-blocks defined')

    ! check if number of blocks is <= 1023
    FD4_L1_ASSERT(size(blx)>1023,'x-blocks must be <= 1023')
    FD4_L1_ASSERT(size(bly)>1023,'y-blocks must be <= 1023')
    FD4_L1_ASSERT(size(blz)>1023,'z-blocks must be <= 1023')
       
    ! check block definition
    do i=1,size(blx)-1
      FD4_L1_ASSERT(blx(i)>=blx(i+1),'incorrect x-blocks')
    end do

    do i=1,size(bly)-1
      FD4_L1_ASSERT(bly(i)>=bly(i+1),'incorrect y-blocks')
    end do

    do i=1,size(blz)-1
      FD4_L1_ASSERT(blz(i)>=blz(i+1),'incorrect z-blocks')
    end do
    
    ! check if number of variables is > 0
    FD4_L1_ASSERT(size(vars)<1,'no variables defined')

    ! check if vartab is valid
    call fd4_vartab_check(vars, ierr)
    FD4_L1_ASSERT(ierr/=0,'invalid variable table')


    call MPI_Comm_dup(mpi_comm,domain%mpi%comm,ierr)
    if(ierr/=MPI_SUCCESS) then
      FD4_L1_ERROR('MPI_Comm_dup failed, err ='//i2s(ierr))
      return
    end if

    call MPI_Comm_size(mpi_comm,domain%mpi%nprocs,ierr)
    call MPI_Comm_rank(mpi_comm,domain%mpi%rank,ierr) 
    
    ! (dummy) allocate the helpers for MPI_Allgatherv
    allocate(domain%mpi%recvcnt(0:domain%mpi%nprocs-1), stat=ierr )
    FD4_L1_ASSERT(ierr/=0,'allocate domain%mpi%recvcnt')
    allocate(domain%mpi%recvdis(0:domain%mpi%nprocs-1), stat=ierr )
    FD4_L1_ASSERT(ierr/=0,'allocate domain%mpi%recvdis')
    allocate(domain%mpi%senddis(0:0), stat=ierr )
    FD4_L1_ASSERT(ierr/=0,'allocate domain%mpi%senddis')

    ! dummy allocate the real(r4k) exchange arrays
    allocate(domain%mpi%recvreal(1), stat=ierr )
    FD4_L1_ASSERT(ierr/=0,'allocate domain%mpi%recvreal')
    allocate(domain%mpi%sendreal(1), stat=ierr )
    FD4_L1_ASSERT(ierr/=0,'allocate domain%mpi%sendreal')

    ! allocate the array of requests
    call stack_create(domain%mpi%requests, 16)
    
    ! dummy allocate the array of statuses
    allocate(domain%mpi%statuses(MPI_STATUS_SIZE,1), stat=ierr )
    FD4_L1_ASSERT(ierr/=0,'allocate domain%mpi%statuses')

    ! allocate marked blocks stack
    call stack_create(domain%markedblocks,1)

    ! allocate and copy vartab
    allocate( domain%vartab(1:size(vars)), stat=ierr )
    FD4_L1_ASSERT(ierr/=0,'allocate domain%vartab')
    domain%vartab(:) = vars(:)
    
    ! create varinfo
    call fd4_vartab_create_varinfo(domain%vartab, domain%varinfo, ierr)
    FD4_L3_ASSERT(ierr/=0,'fd4_vartab_create_varinfo failed')

    ! set number of blocks
    domain%blnum(1) = size(blx)-1
    domain%blnum(2) = size(bly)-1
    domain%blnum(3) = size(blz)-1
    
    ! allocate and copy block definitions
    allocate(domain%d(1)%bl(size(blx)), stat=ierr)
    FD4_L1_ASSERT(ierr/=0,'allocate domain%d(1)%bl')
    domain%d(1)%bl(:) = blx(:)

    allocate(domain%d(2)%bl(size(bly)), stat=ierr)
    FD4_L1_ASSERT(ierr/=0,'allocate domain%d(2)%bl')
    domain%d(2)%bl(:) = bly(:)

    allocate(domain%d(3)%bl(size(blz)), stat=ierr)
    FD4_L1_ASSERT(ierr/=0,'allocate domain%d(3)%bl')
    domain%d(3)%bl(:) = blz(:)

    ! allocate array of b%sempty of all blocks (allocated with boundary)
    if(domain%varinfo%vthres_mode) then
      allocate(domain%empties(0:domain%blnum(1)+1,0:domain%blnum(2)+1,0:domain%blnum(3)+1), stat=ierr)
      FD4_L1_ASSERT(ierr/=0,'allocate domain%empties')
      domain%empties = 0_i4k
    else
      domain%empties => NULL()
    end if

    ! init load balance history stuff
    domain%balance%lbtimes = -100.0_r4k
    domain%balance%curlbtime = 1
    domain%balance%globalize_lbtime = .false.
    domain%balance%lbbalance = -100.0_r4k
    domain%balance%curlbbalance = 1
    domain%balance%part_changed = .false.

    ! check nghosts for each dimension
    do dim=1,3
      ! find minimal block extent
      minext = huge(minext)
      do bl=1,domain%blnum(dim)
        minext = MIN( minext, domain%d(dim)%bl(bl+1) - domain%d(dim)%bl(bl) )
      end do
      ! check if nghosts is <= minimal block extent in this dimension
      !TODO: abandon this limitation
      FD4_L1_ASSERT(nghosts(dim)>minext,'nghosts > smallest block')
      ! check if nghosts is >= 0
      FD4_L1_ASSERT(nghosts(dim)<0,'nghosts < 0')
    end do

    domain%nghosts(:) = nghosts(:)

    ! Periodic boundary conditions are defined for each dimension separately
    domain%periodic(:) = periodic(:)
    
    ! allocate "3D block position" -> "MPI rank" map
    allocate( domain%blkmap(0:domain%blnum(1)+1,0:domain%blnum(2)+1,0:domain%blnum(3)+1,2), stat=ierr )
    FD4_L1_ASSERT(ierr/=0,'allocate domain%blkmap')
    domain%blkmap(:,:,:,:) = -1
    domain%blkmap_now = 1
    domain%blkmap_last = 2

    ! set domain bounds
    do i=1,3
      domain%dbnd(i,:) = (/ domain%d(i)%bl(1), domain%d(i)%bl(domain%blnum(i)+1)-1 /)
    end do
    
    ! calculate level of SFC
    domain%sfclevel = log2i(maxval(domain%blnum))
    
    domain%partition_id = 0_i8k
    domain%partition_called = 0_i8k
    domain%blocks_are_marked = .false.
    
    ! find min and max size of blocks in each dimension
    do dim=1,3
      domain%mmbsz(dim,1)=domain%d(dim)%bl(domain%blnum(dim)+1) - domain%d(dim)%bl(1)
      domain%mmbsz(dim,2)=1
      do i=1,domain%blnum(dim)
        ! number of grid cells in dimension dim
        bl = domain%d(dim)%bl(i+1) - domain%d(dim)%bl(i)
        ! search for min and max size in this dimension
        if( bl < domain%mmbsz(dim,1) ) domain%mmbsz(dim,1) = bl     ! min
        if( bl > domain%mmbsz(dim,2) ) domain%mmbsz(dim,2) = bl     ! max
      end do      
    end do
    
    ! reset spatial union of bounds of all local blocks
    call section_reset(domain%blockunion, domain%dbnd)
    

    !! allocate block pools
    
    ! number of blocks for each block size
    bbnd = domain%mmbsz
    allocate( nbsz( bbnd(1,1):bbnd(1,2), bbnd(2,1):bbnd(2,2), bbnd(3,1):bbnd(3,2) ) )
    nbsz = 0
    ! allocate block pools
    allocate( domain%pool( bbnd(1,1):bbnd(1,2), bbnd(2,1):bbnd(2,2), bbnd(3,1):bbnd(3,2) ) )

#ifndef FD4_DOMAIN_NO_POOL
    ext = 0
    ! loop over all blocks to count the number of blocks for each possible block size
    do z=1,domain%blnum(3)
      do y=1,domain%blnum(2)
        do x=1,domain%blnum(1)
          ext = fd4_domain_blext_ijk(domain, x, y, z)
          nbsz(ext(1),ext(2),ext(3)) = nbsz(ext(1),ext(2),ext(3)) + 1
        end do
      end do
    end do
#endif

    nstacks = 0
    totstacksize = 0

    ! loop over all possible block sizes to set the size of each pool:
    !    2 * max number of blocks / number of ranks
    do z=bbnd(3,1),bbnd(3,2)
      do y=bbnd(2,1),bbnd(2,2)
        do x=bbnd(1,1),bbnd(1,2)
          bl = nbsz(x,y,z)
          if(bl>0) then
            bl = 2 * MAX( MIN(2,bl), bl/domain%mpi%nprocs )
            nstacks = nstacks + 1
            totstacksize = totstacksize + bl
            domain%pool(x,y,z)%len = 0
            domain%pool(x,y,z)%start => null()
            domain%pool(x,y,z)%maxlen = bl
            if(domain%mpi%rank==0) then
              FD4_L5_INFO('pool '//iii2s(x,y,z)//'  size: '//i2s(bl))
            end if
          end if
        end do
      end do
    end do

    !! allocate ghost block pools

    ! number of ghost blocks for each possible ghost block size
    gbbnd = domain%mmbsz
    where(domain%nghosts>0) gbbnd(:,1) = domain%nghosts
    allocate( ngbsz( gbbnd(1,1):gbbnd(1,2), gbbnd(2,1):gbbnd(2,2), gbbnd(3,1):gbbnd(3,2) ) )
    ngbsz = 0
    ! gpmap (ghost pool map) : maps ghost block extent to position in ghost block pool
    !  this leads to a smaller number of empty ghost block pools
    !  add this to the extent of a ghost block, but not to a dimension which has the extent = nghosts!
    domain%gpmap(1:3) = -domain%mmbsz(1:3,1)+domain%nghosts(1:3)+1
    ! allocate ghost block pools
    allocate( domain%gpool( gbbnd(1,1):gbbnd(1,2)+domain%gpmap(1), &
                            gbbnd(2,1):gbbnd(2,2)+domain%gpmap(2), &
                            gbbnd(3,1):gbbnd(3,2)+domain%gpmap(3)    ) )
    
#ifndef FD4_DOMAIN_NO_POOL
    ! 2 * (number of procs) ^ (1/3)
    nprocs13 = 2. * MAX( 1., REAL(domain%mpi%nprocs)**(1./3.) )
    ! estimate the number of ghost blocks for each dimension
    if(nghosts(1)>0) then
      ngbsz(nghosts(1),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) = &
      ngbsz(nghosts(1),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) + INT( REAL(SUM(nbsz(:,:,:), DIM=1))**(2./3.) * nprocs13 )
    end if
    if(nghosts(2)>0) then
      ngbsz(bbnd(1,1):bbnd(1,2),nghosts(2),bbnd(3,1):bbnd(3,2)) = &
      ngbsz(bbnd(1,1):bbnd(1,2),nghosts(2),bbnd(3,1):bbnd(3,2)) + INT( REAL(SUM(nbsz(:,:,:), DIM=2))**(2./3.) * nprocs13 )
    end if
    if(nghosts(3)>0) then
      ngbsz(bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),nghosts(3)) = &
      ngbsz(bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),nghosts(3)) + INT( REAL(SUM(nbsz(:,:,:), DIM=3))**(2./3.) * nprocs13 )
    end if
    ! just to ensure there's enough space, we allocate space for twice of the estimated max
    ngbsz = ngbsz * 2
#endif
    
    ! loop over all possible block sizes to set the size of each pool:
    !   2 * ghost blocks of this size / number of ranks
    do z=gbbnd(3,1),gbbnd(3,2)
      do y=gbbnd(2,1),gbbnd(2,2)
        do x=gbbnd(1,1),gbbnd(1,2)
          bl = ngbsz(x,y,z)
          if(bl>0) then
            bl = 2 * MAX( MIN(2,bl), bl/domain%mpi%nprocs )
            nstacks = nstacks + 1
            totstacksize = totstacksize + bl
            ext(1) = x ; ext(2) = y ; ext(3) = z
            where(ext>domain%nghosts) ext = ext + domain%gpmap
            if( ANY(ext<lbound(domain%gpool)) .or. ANY(ext>ubound(domain%gpool)) ) then
              FD4_L1_ERROR('error in gpmap: out of bounds')
            end if
            if(domain%gpool(ext(1),ext(2),ext(3))%maxlen>0) then
              FD4_L1_ERROR('error in gpmap: maxlen>0')
            end if
            domain%gpool(ext(1),ext(2),ext(3))%len = 0
            domain%gpool(ext(1),ext(2),ext(3))%start => null()
            domain%gpool(ext(1),ext(2),ext(3))%maxlen = bl
            if(domain%mpi%rank==0) then
              FD4_L5_INFO('gpool '//iii2s(x,y,z)//' at '//p2s(ext)//'  size: '//i2s(bl))
            end if
          end if
        end do
      end do
    end do
    
    deallocate(nbsz)
    deallocate(ngbsz)
    
    
    !! set optimal params for encoding the index for the binary tree of MPI types for coupling
    !!
    !!                   |  1  |   2  |  3  4  5 |  6  7  8  |  9 10 11  |  12  13  14 |
    !! cplblkenc(1:14) = | var | step | ext(1:3) | bnd0(1:3) | bnd1(1:3) | minext(1:3) |
    !!                    -----------------------------------------------
    !!                               number of bits required for
    !!
    !! Clearly, this has limits. When there are many time steps, variables, and large blocks,
    !! 64 bit are not enough. Realistic limit is block size of 2**8 = 256 in each dim.
    
    ! bits for set bnd0 and bnd1
    do dim=1,3
      domain%cplblkenc(8+dim) = log2i( domain%mmbsz(dim,2) )
      domain%cplblkenc(5+dim) = domain%cplblkenc(8+dim)
    end do
    ! set minext
    domain%cplblkenc(12:14) = domain%mmbsz(1:3,1)
    ! set bits for ext
    ! additional "+ 1" because of face variables!
    do dim=1,3
      domain%cplblkenc(2+dim) = log2i( domain%mmbsz(dim,2) - domain%mmbsz(dim,1) + 1 + 1 )
    end do
    ! set bits for var and step
    domain%cplblkenc(1) = log2i( size(domain%vartab) )
    domain%cplblkenc(2) = log2i( domain%varinfo%maxnsteps )
    if(SUM(domain%cplblkenc(1:11))>64) then
      FD4_L1_ERROR('cannot encode couple blocks in 64bit')
    end if
    ! create displacements from the bit numbers
    do i=1,10
      domain%cplblkenc(i) = SUM(domain%cplblkenc(i+1:11))
    end do
    domain%cplblkenc(11) = 0


    ! create statistics entries for load balancing
    call fd4_stat_add(domain%stat,'Balance check us'  ,domain%balance%stat_check_us)
    call fd4_stat_add(domain%stat,'Balance part  us'  ,domain%balance%stat_part_us)
    call fd4_stat_add(domain%stat,'Balance mig   us'  ,domain%balance%stat_mig_us)
    call fd4_stat_add(domain%stat,'Balance recv Blocks',domain%balance%stat_migblock)
    call fd4_stat_add(domain%stat,'Balance recv Bytes' ,domain%balance%stat_migbytes)
    domain%balance%stat_binsteps = -1 ! will be created in fd4_part_sfc if necessary
    ! create statistics entries for block allocation/deallocation
    call fd4_stat_add(domain%stat,'Domain block alloc',domain%stat_bl_alloc)
    call fd4_stat_add(domain%stat,'Domain block free' ,domain%stat_bl_free)
    call fd4_stat_add(domain%stat,'Domain ghost alloc',domain%stat_gb_alloc)
    call fd4_stat_add(domain%stat,'Domain ghost free' ,domain%stat_gb_free)
    
    ! initialize timing
    call gettime_init()

    ! initialize mapping of 3D to 2D for 2D SFC partitioning
    if(ANY(domain%blnum==1)) then
      if(domain%blnum(1)==1) then
        domain%balance%map2d = (/2,3/)
      elseif(domain%blnum(2)==1) then
        domain%balance%map2d = (/1,3/)
      else
        domain%balance%map2d = (/1,2/)
      end if
    else
      domain%balance%map2d = -1
    end if
    
    ! print some info
    if(FD4_VERBOSE_LEVEL >= 2 .and. domain%mpi%rank == 0) then
    
      write (fd4_out,'(A,A)') FD4_LINEHD,'created new fd4_domain:'
      write (fd4_out,'(A,A)') FD4_LINEHD,'dim   start     end  blocks   blksz  ghosts  per.bd'
      do i=1,3
        write (fd4_out,'(A,A,I8,I8,I8,F8.2,I8,L8)')  FD4_LINEHD,dimstr(i),   &
              domain%dbnd(i,1),domain%dbnd(i,2),        domain%blnum(i),  &
             (domain%dbnd(i,2)-domain%dbnd(i,1)+1)/REAL(domain%blnum(i)), &
              domain%nghosts(i), domain%periodic(i)
      end do
      
      write (fd4_out,'(A,A)') FD4_LINEHD,'max. number of blocks:   '//l2s(INT(PRODUCT(domain%blnum(1:3)),r8k))
      write (fd4_out,'(A,A)') FD4_LINEHD,'SFC level:               '//i2s(domain%sfclevel)
      write (fd4_out,'(A,A)') FD4_LINEHD,'number of MPI processes: '//i2s(domain%mpi%nprocs)
      write (fd4_out,'(A,A)') FD4_LINEHD,'block pool lists:        '//i2s(nstacks)
      write (fd4_out,'(A,A)') FD4_LINEHD,'block pool max. size:    '//i2s(totstacksize)
      write (fd4_out,'(A,A,L1)') FD4_LINEHD,'adaptive block mode:     ',domain%varinfo%vthres_mode

      write (fd4_out,'(A,A)') FD4_LINEHD,'variable table:'
      write (fd4_out,'(A,A)') FD4_LINEHD,' id  nbins  nsteps  dyn     vnull  threshld  face  name'
      do i=1,size(domain%vartab)
        v => domain%vartab(i)
        if(v%vthres == FD4_NOTHRES) then
          vthres='         -'
        else
          write(vthres,'(A,ES8.1)') '  ',v%vthres
        end if
        write (fd4_out,'(A,I3,I7,I8,L5,A,ES8.1,A,A,A,A,A)') FD4_LINEHD,i,v%nbins, &
               v%nsteps,v%dynamic,'  ',v%vnull,vthres,         &
               '   ',dimstr(v%facevar),                        &
               '  ',v%name(1:len_trim(v%name))
      end do

    end if

#ifdef FD4_COUPLE_DEBUG
    if(domain%mpi%rank == 0) then
      write (fd4_out,'(A,A,2(I4))') FD4_LINEHD,'cplblkenc (var, step) = ', domain%cplblkenc(1:2)
      write (fd4_out,'(A,A,3(I4))') FD4_LINEHD,'cplblkenc (ext)       = ', domain%cplblkenc(3:5)
      write (fd4_out,'(A,A,3(I4))') FD4_LINEHD,'cplblkenc (bnd0)      = ', domain%cplblkenc(6:8)
      write (fd4_out,'(A,A,3(I4))') FD4_LINEHD,'cplblkenc (bnd1)      = ', domain%cplblkenc(9:11)
      write (fd4_out,'(A,A,3(I4))') FD4_LINEHD,'cplblkenc (minext)    = ', domain%cplblkenc(12:14)
    end if
#endif

    domain%init = .true.
    err = 0
    
  
  end subroutine fd4_domain_create_specific


  !! Initialize the decomposed management of meta data for coupling.
  !!
  !! This subroutine must be called before creating couple contexts for $domain.
  !! If it is not called, no decomposed management is applied.
  !! Decomposed management is not implemented for adaptive block mode.
  !!
  !! With $maxext you can set the maximum extent of couple arrays in number of grid cells per dimension
  !! (without extension by face variables).
  !! If any value in $maxext is <=0, decomposition is disabled.
  !!
  subroutine fd4_domain_init_metadata(domain, maxext, err)
    implicit none
    type(fd4_domain), intent(inout) :: domain    !! the domain
    integer(i_k)    , intent(in)    :: maxext(3) !! max extent of couple arrays
    integer(i_k)    , intent(out)   :: err       !! error status: 0...ok

    integer(i_k) :: mdsdext(3) ! meta data subdomain size
    integer(i_k) :: dim, i, ii, lc, dext(3)

    FD4_L1_ASSERT(.not.domain%init,'domain not initialized')

    dext(1:3) = domain%dbnd(1:3,2) - domain%dbnd(1:3,1) + 1 ! domain extent

    if(domain%varinfo%vthres_mode .or. ANY(maxext<=0)) then
      ! one large mdsd will be used in adaptive block mode
      ! (no decomposed management)
      mdsdext = domain%blnum
    else
      ! guess a good value for meta data subdomain size from given mexext
      ! (the number of mdsd's should be ~< number of processes,
      !  extent of the mdsd's should be a power of two)
      mdsdext = 1
      ! get the max extent of a couple array as number of blocks for each dimension
      do dim=1,3
        do i=1,domain%blnum(dim)
          lc = domain%d(dim)%bl(i+1) - 1 ! last grid cell of this block in dim
          lc = MIN(lc + maxext(dim), dext(dim))  ! not "+maxext(dim)-1" because of face variables
          do ii=i+1,domain%blnum(dim)
            if(lc <= domain%d(dim)%bl(ii+1) - 1) then
              mdsdext(dim) = MAX(mdsdext(dim), ii-i)
              exit
            end if
          end do
        end do
        if(mdsdext(dim) > domain%blnum(dim)/2) then
          ! enlarge to whole domain
          mdsdext(dim) = domain%blnum(dim)
        else
          ! enlarge to next power of 2
          mdsdext(dim) = 2**log2i(mdsdext(dim))
        end if
      end do
      ! enlarge size of mdsd's until the number of mdsd's is <= number of processes
      dim = 1
      do while( PRODUCT(REAL(domain%blnum)/REAL(mdsdext)) > REAL(domain%mpi%nprocs) )
        if(mdsdext(dim)<domain%blnum(dim)) then
          mdsdext(dim) = MIN(mdsdext(dim)*2, domain%blnum(dim))
          !print *,'enlarge ',dim,' :',domain%blnum/mdsdext
        end if
        dim = MOD(dim,3) + 1
      end do
    end if

    call fd4_metadata_init(domain%md, domain%blnum, mdsdext, err)
    FD4_L1_ASSERT(err/=0,'cannot initialize metadata') 
    
    if(domain%mpi%rank==0 .and. ANY(domain%md%mdsdnum>1)) then
      FD4_L2_INFO('number of mdsds: '//p2s(domain%md%mdsdnum))
    end if

    call fd4_metadata_full_allocate(domain%md)
    
    err = 0
    
  end subroutine fd4_domain_init_metadata
  
  
  !! Optimize metadata for coupling for current local blocks.
  !! 
  !! Once this subroutine has been called, the meta data will be optimized at each repartitioning.
  !! You must first initialize decomposed management of meta date with $fd4_domain_init_metadata.
  !!
  subroutine fd4_domain_adapt_metadata(domain)
    type(fd4_domain), intent(inout) :: domain    !! the domain
  
    domain%md%adapt = domain%md%initialized .and. ( .not. domain%md%single_mdsd)
    if(domain%md%adapt) then
      call fd4_metadata_adapt(domain%md, domain%blocks)
    end if
    
  end subroutine fd4_domain_adapt_metadata


  !! Collect global min/max/avg/sum values of the FD4 statistics and let rank $opt_root
  !! print the statistics to unit $out.
  !! The default unit is stdout. The default root is rank 0.
  !!
  subroutine fd4_domain_dump_stats(domain, opt_root, opt_out)
    implicit none
    type(fd4_domain), intent(inout) :: domain    !! the domain
    integer, intent(in), optional :: opt_root    !! root MPI process that prints statistics
    integer, intent(in), optional :: opt_out     !! unit for output
    
    call fd4_stat_dump(domain%stat,domain%mpi, opt_root, opt_out)
  
  end subroutine fd4_domain_dump_stats


  !! Delete all blocks of a domain.
  subroutine fd4_domain_clear(domain)
    implicit none
    type(fd4_domain), intent(inout) :: domain !! the domain
    
    type(fd4_block), pointer :: b
    integer(i_k) :: gtype
    integer(i_k) :: x, y, z
    
    do gtype=-1,0
      ! deallocate all boundary and comm ghost blocks
      do while(associated(domain%glist(gtype)%start))
        b => domain%glist(gtype)%start
        call fd4_ghostlist_remove(domain%glist(gtype), b)
        call fd4_block_deallocate(b)
      end do
    end do
    
    ! deallocate all blocks
    call rbtree_fd4_block_clear(domain%blocks,fd4_block_deallocate)

    ! deallocate blocks from the block pools
    do z=lbound(domain%pool,3),ubound(domain%pool,3)
      do y=lbound(domain%pool,2),ubound(domain%pool,2)
        do x=lbound(domain%pool,1),ubound(domain%pool,1)
          ! deallocate the blocks which still reside in the pool
          do while(associated(domain%pool(x,y,z)%start))
            b => domain%pool(x,y,z)%start
            domain%pool(x,y,z)%start => b%neigh(1,1)%l
            call fd4_block_deallocate(b)
          end do
          domain%pool(x,y,z)%len = 0
        end do
      end do
    end do    

    ! deallocate blocks from the ghost block pools
    do z=lbound(domain%gpool,3),ubound(domain%gpool,3)
      do y=lbound(domain%gpool,2),ubound(domain%gpool,2)
        do x=lbound(domain%gpool,1),ubound(domain%gpool,1)
          ! deallocate the blocks which still reside in the pool
          do while(associated(domain%gpool(x,y,z)%start))
            b => domain%gpool(x,y,z)%start
            domain%gpool(x,y,z)%start => b%neigh(1,1)%l
            call fd4_block_deallocate(b)
          end do
          domain%gpool(x,y,z)%len = 0
        end do
      end do
    end do
    
    call fd4_stat_reset(domain%stat)
    
  end subroutine fd4_domain_clear



  !! Delete a domain.
  !!
  !! Calls $fd4_domain_dump_stats and deallocates all data associated with domain.
  subroutine fd4_domain_delete(domain, opt_stats)

    implicit none

    type(fd4_domain), intent(inout) :: domain  !! the domain
    logical, intent(in), optional :: opt_stats !! dump FD4 statistics to stdout? default is yes.
    
    integer(i_k) :: i, err, datatype
    type(rbtree_int8_iter) :: int8iter
    logical :: stats
    
    stats = .true.
    if(present(opt_stats)) stats = opt_stats

    if(FD4_VERBOSE_LEVEL >= 2 .and. .not. domain%init) then
       write(fd4_out,'(A,A)') FD4_LINEHD,'Warning: deleting domain which is already empty'
    end if
    
    ! dump statistics
    if(stats .and. domain%stat_enabled) call fd4_domain_dump_stats(domain)

    ! delete all blocks and their data
    call fd4_domain_clear(domain)
    
    ! delete the statistics
    call fd4_stat_free(domain%stat)

    ! free meta data subdomains
    call fd4_metadata_finalize(domain%md)

    domain%dbnd(:,:)  = 0_i_k
    domain%blnum(:)   = 0_i_k
    
    if(associated(domain%vartab)) deallocate(domain%vartab)
    nullify(domain%vartab)
    
    if(associated(domain%varinfo%fields)) deallocate(domain%varinfo%fields)
    nullify(domain%varinfo%fields)
    
    deallocate(domain%mpi%recvcnt)
    deallocate(domain%mpi%recvdis)
    deallocate(domain%mpi%senddis)
    deallocate(domain%mpi%recvreal)
    deallocate(domain%mpi%sendreal)
    deallocate(domain%mpi%statuses)
    if(associated(domain%empties)) deallocate(domain%empties)
    if(associated(domain%balance%blkweights))    deallocate(domain%balance%blkweights)
    if(associated(domain%balance%blkweights_ps)) deallocate(domain%balance%blkweights_ps)
    
    if(associated(domain%balance%rnk))  deallocate(domain%balance%rnk)
    if(associated(domain%balance%hidx)) deallocate(domain%balance%hidx)

    call stack3_delete(domain%balance%allblocks)

    call stack_delete(domain%mpi%requests)
    call stack_delete(domain%markedblocks)

    do i=1,3
      if(associated(domain%d(i)%bl)) deallocate(domain%d(i)%bl)
    end do

    if(associated(domain%blkmap)) deallocate(domain%blkmap)
    nullify(domain%blkmap)

    ! deallocate block pools
    deallocate(domain%pool)
    deallocate(domain%gpool)
    
    ! clear and deallocate cpltypestore
    datatype = rbtree_int8_iter_first(int8iter,domain%cpltypestore)
    do while(associated(int8iter%cur))
      call MPI_Type_free(datatype, err)
      datatype = rbtree_int8_iter_next(int8iter)
    end do
    call rbtree_int8_clear(domain%cpltypestore)
    
    ! clear memory of meta data management for coupling
    ! (if all couple contexts have been removed, no memory should be left to remove)
    call fd4_metadata_clear(domain%md)
    
    ! free communicator
    CALL MPI_Comm_free(domain%mpi%comm,err)
    if(err/=MPI_SUCCESS) then
      FD4_L1_ERROR('MPI_Comm_free failed, err ='//i2s(err))
    end if
    
    domain%init = .false.
  
  end subroutine fd4_domain_delete
  

end module fd4_domain_mod
