! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!             Globad definitions and helper routines           !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Global definitions for FD4 and some helper routines. 
!!
!! Proprocessor options:
!!  - $NO_REC_IO: Use this for compilers which do not support recursive I/O 
!!    (e.g. PGI).
!!
!! The subroutines $fd4_lerr and $fd4_perr should not be
!! called directly, but used through the error handling interface
!! defined in $fd.$h. Following *verbose levels for error handling*
!! are defined:
!!>  L1 =      output of basic and cheap error checks (default)
!!   L2 = L1 + some basic information
!!   L3 = L2 + expensive error checks (will hurt performance)
!!<  L4 = L3 + verbose mode (tons of output)
!!
!! L1 error checks will be always carried through (due to semantics
!! of subroutines), but output of these error messages will only be
!! enabled if $FD4_VERBOSE_LEVEL >= 1.
!! L3 error checks are only arried through if L3 is enabled.
!!
!! The default verbose level is L1. Set the level with
!!v -DFD4_VERBOSE_LEVEL=[1|2|3|4]
!! when compiling the source file(s).
!!
!! The error handling interface in $fd.$h defines the following preprocessor macros:
!!> FD4_L1_ASSERT(C,MSG)
!!  FD4_L1_ERROR(MSG)
!!  FD4_L1_WARN(MSG)
!!  FD4_L3_ASSERT(C,MSG)
!!< FD4_L4_INFO(MSG)
!! $C is a logical expression, which should be true when an error condition occured.
!! $MSG is a character describing the error/warning/info.
!!
!! @author Matthias Lieber
module fd4_globaldef_mod

  use kinds_mod
  implicit none
  
  public

  !! MPI_COMM_WORLD rank
  integer :: fd4_wrnk

  !! head/flag for fd4 output lines
  character(11) :: FD4_LINEHD

  !! logical unit for stdout
  integer, parameter :: fd4_out = 6
  
  !! mapping from (1,2) to (decrement,increment)
  integer(i_k), parameter :: fd4_incdec(2) = (/-1,1/)
  
  !! constants for specifying (multiple) dimensions
  integer(i_k), parameter :: FD4_X = 1, FD4_Y = 2, FD4_Z = 4
  integer(i_k), parameter :: FD4_XY  = IOR(FD4_X, FD4_Y)
  integer(i_k), parameter :: FD4_XZ  = IOR(FD4_X, FD4_Z)
  integer(i_k), parameter :: FD4_YZ  = IOR(FD4_Y, FD4_Z)
  integer(i_k), parameter :: FD4_XYZ = IOR(FD4_XY,FD4_Z)  
  integer(i_k), parameter :: FD4_DIM(1:3) = (/ FD4_X, FD4_Y, FD4_Z /)

contains


  !! Print the kind values defined in $kinds_mod.
  !! @see kinds_mod
  subroutine fd4_print_kinds
  
    implicit none

    write(fd4_out,'(A,A,I3)') FD4_LINEHD,' 4 byte integer i4k =',i4k
    write(fd4_out,'(A,A,I3)') FD4_LINEHD,' 8 byte integer i8k =',i8k
    write(fd4_out,'(A,A,I3)') FD4_LINEHD,'default integer i_k =',i_k

    write(fd4_out,'(A,A,I3)') FD4_LINEHD,' 4 byte real    r4k =',r4k
    write(fd4_out,'(A,A,I3)') FD4_LINEHD,' 8 byte real    r8k =',r8k
    write(fd4_out,'(A,A,I3)') FD4_LINEHD,'default real    r_k =',r_k

  end subroutine fd4_print_kinds



  !! Print error message upon condition.
  !!
  !!  If $err == .true. print error message and return true,
  !!  else                                      return false.
  logical function fd4_lerr(line, file, err, msg)

    implicit none
    integer, intent(in)      :: line       !! source code line of the error
    character(*), intent(in) :: file       !! source code file of the error
    logical, intent(in)      :: err        !! error condition
    character(*), intent(in) :: msg        !! error message to print, if $err == .true.
    
    fd4_lerr = err
    if(.not. err) return

    call fd4_perr(line, file, msg)
    
  end function fd4_lerr
  

  !! Print error message.
  !!
  subroutine fd4_perr(line, file, msg)

    implicit none
    integer, intent(in)      :: line       !! source code line of the error
    character(*), intent(in) :: file       !! source code file of the error
    character(*), intent(in) :: msg        !! error message to print
    character(len=8) :: linestr

    write(linestr,'(I7)') line
    linestr = adjustl(linestr)
    write(fd4_out,'(A,A,A,A,A,A,A)') &
         FD4_LINEHD,'Error at ',STRIM(file),':',STRIM(linestr), &
         ': ',STRIM(msg)
    
  end subroutine fd4_perr
  


  ! The following functions lead to recursive I/O if used the following way:
  !   write(*,'(A)') 'the number is: '//i2s(number)
  ! Since this is very common in FD4, I wrote routines which don't use
  ! the write statement for compilers not supporting recursive I/O.
  ! Enable these alternatives by defining NO_REC_IO.


  !! convert number $inum to character of length 6
  !!
  function i2s(inum,zero) result(str)
    implicit none
    integer(i_k), intent(in) :: inum         !! number to be converted to a character
    logical, optional, intent(in) :: zero !! fill with leading zeros (to avoid blanks)
    character(6) :: str
#ifdef NO_REC_IO
    integer(i_k) :: i,ic1,ic2,begin,p0,p1
    p0=1
    p1=6   
    ic1=ABS(inum)
    begin=0
    do i=p1,p0,-1
      if(ic1==0) then
        begin=i
        if(inum==0) then
          str(i:i)='0'
          begin=i-1
        end if
        exit
      else
        ic2=ic1/10
        str(i:i)=achar(48+ic1-(ic2*10))      
        ic1=ic2
      end if
    end do
    if(ic1/=0 .or. (inum<0 .and. begin==0) ) then
      str='******'
    else
      if(present(zero)) then
        do i=p0,begin
          str(i:i)='0'
        end do
        if(inum<0) str(p0:p0)='-'
      else
        do i=p0,begin
          str(i:i)=' '
        end do
        if(inum<0) str(begin:begin)='-'
        str = adjustl(str)
      end if
    end if
#else
    str=''
    if(present(zero)) then
      write(str,'(I6.6)') inum
    else
      write(str,'(I6)') inum
      str = adjustl(str)
    endif
#endif
  end function i2s


  !! convert number $inum to character of length 12
  !!
  function i2sl(inum,zero) result(str)
    implicit none
    integer(i_k), intent(in) :: inum         !! number to be converted to a character
    logical, optional, intent(in) :: zero !! fill with leading zeros (to avoid blanks)
    character(12) :: str
#ifdef NO_REC_IO
    integer(i_k) :: i,ic1,ic2,begin,p0,p1
    p0=1
    p1=12   
    ic1=ABS(inum)
    begin=0
    do i=p1,p0,-1
      if(ic1==0) then
        begin=i
        if(inum==0) then
          str(i:i)='0'
          begin=i-1
        end if
        exit
      else
        ic2=ic1/10
        str(i:i)=achar(48+ic1-(ic2*10))      
        ic1=ic2
      end if
    end do
    if(ic1/=0 .or. (inum<0 .and. begin==0) ) then
      str='************'
    else
      if(present(zero)) then
        do i=p0,begin
          str(i:i)='0'
        end do
        if(inum<0) str(p0:p0)='-'
      else
        do i=p0,begin
          str(i:i)=' '
        end do
        if(inum<0) str(begin:begin)='-'
        str = adjustl(str)
      end if
    end if
#else
    str=''
    if(present(zero)) then
      write(str,'(I12.12)') inum
    else
      write(str,'(I12)') inum
      str = adjustl(str)
    endif
#endif
  end function i2sl


  !! convert 8 byte integer $inum to character of length 12
  !!
  function l2s(inum) result(str)
    implicit none
    integer(i8k), intent(in) :: inum  !! number to be converted to a character
    character(12) :: str
#ifdef NO_REC_IO
    integer(i_k) :: i, begin, p0, p1
    integer(i8k) :: ic1,ic2
    p0=1
    p1=12
    ic1=ABS(inum)
    begin=0
    do i=p1,p0,-1
      if(ic1==0) then
        begin=i
        if(inum==0) then
          str(i:i)='0'
          begin=i-1
        end if
        exit
      else
        ic2=ic1/10
        str(i:i)=achar(48+ic1-(ic2*10))      
        ic1=ic2
      end if
    end do
    if(ic1/=0 .or. (inum<0 .and. begin==0) ) then
      str='************'
    else
      do i=p0,begin
        str(i:i)=' '
      end do
      if(inum<0) str(begin:begin)='-'
      str = adjustl(str)
    end if
#else
    str=''
    write(str,'(I12)') inum
    str = adjustl(str)
#endif
  end function l2s


  !! convert position $pos to character of length 20
  !!
  function p2s(p) result(str)
    implicit none
    integer(i_k), intent(in) :: p(3)  !! position to be converted to a character
    character(20) :: str
#ifdef NO_REC_IO
    str(1:1)   = '('
    str(2:7)   = i2s(p(1))
    str(8:13)  = i2s(p(2))
    str(14:19) = i2s(p(3))
    str(20:20) = ')'
#else
    str=''
    write(str,'(A,I6,I6,I6,A)') '(',p(1),p(2),p(3),')'
#endif
  end function p2s

  !! convert integers $inum1, $inum2, $inum3 to character of length 18
  !!
  function iii2s(inum1,inum2,inum3) result(str)
    implicit none
    integer(i_k), intent(in) :: inum1    !! integer to be converted to a character
    integer(i_k), intent(in) :: inum2    !! integer to be converted to a character
    integer(i_k), intent(in) :: inum3    !! integer to be converted to a character
    character(18) :: str
#ifdef NO_REC_IO
    str(1:6)   = i2s(inum1)
    str(7:12)  = i2s(inum2)
    str(13:18) = i2s(inum3)
#else
    str=''
    write(str,'(I6,I6,I6)') inum1,inum2,inum3
#endif
  end function iii2s

  
end module fd4_globaldef_mod
