! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!               Space-filling curve partitioning               !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Space-filling curve partitioning of an $fd4_domain.
!! Supports Hilbert and Morton SFC.
!! One-dimensional partitioning algorithms from $fd4_part_1d_mod are used.
!!
!! @see hilbert_mod
!! @see fd4_part_1d_mod
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_part_sfc_mod

  use fd4_globaldef_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_part_1d_mod
  use fd4_mpi_mod
  use fd4_stat_mod
  use hilbert_mod
  use hilbert2d_mod
  use morton_mod
  use morton2d_mod
  use mrgrnk_mod
  use stack_mod
  use stack3_mod
  use rbtree_fd4_block_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  integer(i_k), parameter, public :: FD4_PART_SFC_SIMPLE = 1
  integer(i_k), parameter, public :: FD4_PART_SFC_BINSRCH = 2
  integer(i_k), parameter, public :: FD4_PART_SFC_PARALLEL = 4
  
  integer(i_k), parameter, public :: FD4_PART_SFC_HILBERT = 128
  integer(i_k), parameter, public :: FD4_PART_SFC_MORTON  = 256
  
  ! public methods
  public :: fd4_part_sfc_noweight
  public :: fd4_part_sfc_weight
  public :: fd4_part_sfc_weight2
  
  !! general iterator over SFC
  type part_sfc_iter
    integer(i_k) :: p(3) = -1    ! spatial position
    integer(i_k) :: idx = -1     ! position on curve
    integer(i_k) :: sfctype = -1
    integer(i_k) :: map2d(2) = 1
    type(hilbert_iter) :: hiter
    type(hilbert2d_iter) :: h2diter
    type(morton_iter) :: miter
    type(morton2d_iter) :: m2diter
  end type part_sfc_iter
  
contains  

  include 'framework/fd4_domain.incf'
  include 'util/stack.incf'
  include 'util/stack3.incf'
  include 'util/math.incf'


  !! Create a new partition using block position. Considers block weights from domain%balance%blkweights.
  !!
  !! This version only works in non-adaptive block mode and when the SFC type is never changed.
  !!
  !! Optimizations in comparison to $fd4_part_sfc_weight:
  !!  - sorting of the weight arrays is not necessary
  !!  - $domain%$balance%$blkweights is not required
  !!  - $domain%$balance%$rnk and $domain%$balance%$hidx are not required
  !!
  !! The prefix sum is still serial.
  !!
  subroutine fd4_part_sfc_weight2(domain, local_blocks, method, sfctype, &
                                  balance, err, opt_binsteps, opt_accuracy)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    type(stack3)    , intent(inout) :: local_blocks                   !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(in)    :: method                         !! 1D partitioning method to use
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    real(r4k)       , intent(out)   :: balance                        !! estimated balance of new partitioning
    integer(i_k)    , intent(out)   :: err                            !! error status: 0...ok
    integer(i_k),optional,intent(in):: opt_binsteps                   !! max. number of steps for binary search methods, default: 100
    !! target ratio from best possible load balance for binary steps estimation, *must be <= 1.0*, default: 0.99
    real(r8k),   optional,intent(in):: opt_accuracy
    target :: domain

    integer :: rank, nprocs, nblocks
    integer(i_k) :: ib, crank, pos(3), pstart, pend
    real(r8k) :: maxweight, maxload
    real(r8k), pointer :: blkweights_ps(:)     ! prefix sum of sfc sorted block weights
    integer(i_k) :: partition(0:domain%mpi%nprocs) ! partition vector (start indices)
    integer(i_k) :: binsteps, reqsteps
    real(r8k) :: accuracy, eps
    type(fd4_iter) :: iter
    type(fd4_mpi), pointer :: mpi
    type(part_sfc_iter) :: sfciter

    err = -1

    FD4_L1_ASSERT(domain%varinfo%vthres_mode,'not allowed in adaptive block mode')
    if(domain%blockcount>0 .and. sfctype/=domain%balance%last_sfctype) then
      FD4_L1_ERROR('cannot change SFC type')
    end if
    if(domain%blocks%length>0 .and. domain%balance%sfc_pstart<0) then
      FD4_L1_ERROR('domain%balance%sfc_pstart not set')
    end if

    mpi    => domain%mpi
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    nblocks = PRODUCT(domain%blnum(1:3))
    
    if(.not. associated(domain%balance%blkweights_ps)) then
      allocate(domain%balance%blkweights_ps(0:PRODUCT(domain%blnum(1:3))), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate domain%balance%blkweights_ps')
    end if
    blkweights_ps => domain%balance%blkweights_ps
    

    if(domain%blockcount == nblocks) then
      !! all blocks already allocated
    
      ! allocate array for sending the block weights
      if( size(mpi%sendreal,1) < domain%blocks%length ) then
        deallocate(mpi%sendreal)
        allocate(mpi%sendreal(domain%blocks%length))
      end if
      
      ! get the local block weights and write them to mpi%sendreal (ordered by SFC)
      iter%cur => rbtree_fd4_block_iter_first(iter%biter, domain%blocks)
      do while(associated(iter%cur))
        call fd4_part_sfc_c2i_qsz(domain, sfctype, iter%cur%pos, ib)
        mpi%sendreal(ib-domain%balance%sfc_pstart+1) = iter%cur%weight
        iter%cur => rbtree_fd4_block_iter_next(iter%biter)
      end do

      ! collect the block weights to mpi%recvreal (r4k array)
      call fd4_part_sfc_exch_weights(domain, err, mpi%sendreal, .false.)
      FD4_L1_ASSERT(err/=0,'fd4_part_sfc_exch_weights failed')

      ! apply prefix sum to weight array
      maxweight = 0.0_r8k
      blkweights_ps(0) = 0.0_r8k
      do ib=1,nblocks
        blkweights_ps(ib) = REAL(mpi%recvreal(ib), r8k) ! convert to r8k
        maxweight = MAX(maxweight, blkweights_ps(ib))
        blkweights_ps(ib) = blkweights_ps(ib) + blkweights_ps(ib-1)
      end do

    elseif(domain%blockcount == 0) then
      !! no blocks allocated yet

      ! loop over all (non-allocated) blocks and fill blkweights_ps (prefix sum of weight array)
      maxweight = 0.0_r8k
      blkweights_ps(0) = 0.0_r8k
      do ib=1,nblocks
        call fd4_part_sfc_i2c_qsz(domain, sfctype, ib, pos)
        blkweights_ps(ib) =   REAL( domain%balance%defaultweight, r8k)  &
                            * REAL( PRODUCT( fd4_domain_blext(domain,pos) ), r8k )
        maxweight = MAX(maxweight, blkweights_ps(ib))
        blkweights_ps(ib) = blkweights_ps(ib) + blkweights_ps(ib-1)
      end do
      !if(mpi%rank==0) print *,'#',blkweights_ps

    else
      FD4_L1_ERROR('not allowed in adaptive block mode')
    end if
    
    domain%balance%last_sfctype = sfctype

    if(method==FD4_PART_SFC_BINSRCH .or. method==FD4_PART_SFC_PARALLEL) then
      accuracy = 0.99_r8k
      binsteps = 100
      if(present(opt_accuracy)) accuracy=MAX(MIN(opt_accuracy,1.0_r8k),0.1_r8k)
      if(present(opt_binsteps)) binsteps=MAX(MIN(opt_binsteps,10000),1)
      ! compute the bottleneck error value which is necessary to get
      !  (accuracy * optimal_balance)
      ! as the resulting loadbalance of the 1d partitioning
      !  (eps is 0 if accuracy is 1.0: in this case the methods is exact binsteps is large enough)
      eps = (1.0_r8k-accuracy) * blkweights_ps(nblocks) / (accuracy*nprocs)
      ! this would be the necessary number of binary search steps for the given accuracy:
      !steps = log2i( CEILING( maxweight / eps, i_k ) )
      ! add statistics if necesary
      if(domain%balance%stat_binsteps<0) then
        call fd4_stat_add(domain%stat,'Balance max binsteps'  ,domain%balance%stat_binsteps)
      end if
    end if

    !! call partitioning routine

    if(method==FD4_PART_SFC_BINSRCH) then
      ! good serial method
      call fd4_part_1d_binsrch(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, &
                                       partition, maxload, reqsteps)
      domain%stat%values(domain%balance%stat_binsteps) = MAX(domain%stat%values(domain%balance%stat_binsteps), INT(reqsteps,i8k))
      !
    elseif(method==FD4_PART_SFC_PARALLEL) then    
      ! parallel method
      call fd4_part_1d_parallel(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, &
                                        mpi, partition, maxload, reqsteps, err)
      domain%stat%values(domain%balance%stat_binsteps) = MAX(domain%stat%values(domain%balance%stat_binsteps), INT(reqsteps,i8k))
      !
    else
      ! simple method
      call fd4_part_1d_simple_h1(nblocks, nprocs, blkweights_ps(0:nblocks), partition, maxload)
      !
    end if

    balance = REAL( blkweights_ps(nblocks) / (nprocs * maxload), r4k)

    !! finally apply partition to FD4

    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now

    ! allocate local list of blocks
    pstart = partition(rank)
    pend   = partition(rank+1)-1
    if(pstart<=pend) then
      domain%balance%sfc_pstart = pstart
      domain%balance%localweight = REAL(blkweights_ps(pend) - blkweights_ps(pstart-1),r4k)
      call stack3_create(local_blocks, pend - pstart + 1)
      local_blocks%pos = pend - pstart + 1
    else
      domain%balance%sfc_pstart = -1
      domain%balance%localweight = 0.0
      local_blocks%pos = 0
    end if

    ! iterate over SFC to set block map and local partition
    crank = 0
    call fd4_part_sfc_iter_init(sfciter, domain, sfctype)
    do while(sfciter%idx>0)
      ! advance rank
      do while( crank < mpi%nprocs-1 .and. sfciter%idx >= partition(crank+1) )
        crank = crank + 1
      end do
      ! set block map
      domain%blkmap( sfciter%p(1), sfciter%p(2), sfciter%p(3), domain%blkmap_now ) = crank
      ! set local partition vector
      if(crank == rank) then
        FD4_L3_ASSERT(local_blocks%pos<sfciter%idx-pstart+1,'local blocks')
        local_blocks%a(1:3,sfciter%idx-pstart+1) = sfciter%p(1:3)
      end if
      call fd4_part_sfc_iter_next(sfciter)
    end do


    ! fill statistics
    domain%stat%sfcbnd(1) = MIN(domain%stat%sfcbnd(1), pstart)
    domain%stat%sfcbnd(2) = MAX(domain%stat%sfcbnd(2), pend)
    
    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    call fd4_part_sfc_blkmap_boundary(domain)    
    
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6,E13.4)') FD4_LINEHD,'new partition: ',local_blocks%pos, domain%balance%localweight
#endif
  
    err = 0

  end subroutine fd4_part_sfc_weight2
  

  !! Create a new partition using block position. Considers block weights from domain%balance%blkweights.
  !!
  !! In adaptive block mode $all_blocks may be $NULL() and $all_blocks_len is ignored.
  !!
  !! If $opt_blocks_changed is present and is set to false: $fd4_part_sfc_weight reuses the hilbert indexes
  !! and the ranking of blocks by the hilbert index from the last call!
  !!
  subroutine fd4_part_sfc_weight(domain, all_blocks, all_blocks_len, local_blocks, method, sfctype, &
                                 balance, err, opt_blocks_changed, opt_binsteps, opt_accuracy)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: all_blocks_len                 !! length of $all_blocks(3,:)
    integer(i_k)    , pointer       :: all_blocks(:,:)                !! array of positions of the blocks in the new partion
    type(stack3)    , intent(inout) :: local_blocks                   !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(in)    :: method                         !! 1D partitioning method to use
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    real(r4k)       , intent(out)   :: balance                        !! estimated balance of new partitioning
    integer(i_k)    , intent(out)   :: err                            !! error status: 0...ok
    logical,     optional,intent(in):: opt_blocks_changed             !! tell me if the blocks have changed (created/removed)
    integer(i_k),optional,intent(in):: opt_binsteps                   !! max. number of steps for binary search methods, default: 100
    !! target ratio from best possible load balance for binary steps estimation, *must be <= 1.0*, default: 0.99
    real(r8k),   optional,intent(in):: opt_accuracy
    
    integer :: rank, nprocs, nblocks
    integer(i_k) :: ib, i, j, k, crank, pos(3), pstart, pend
    real(r8k) :: maxweight, weight, maxload
    logical :: blocks_changed
    real(r8k), pointer :: blkweights_ps(:)     ! prefix sum of sfc sorted block weights
    integer(i_k) :: partition(0:domain%mpi%nprocs) ! partition vector (start indices)
    integer(i_k) :: binsteps, reqsteps
    real(r8k) :: accuracy, eps
    
    err = -1
    rank   = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    blocks_changed = .true.
    
    if(domain%varinfo%vthres_mode) then
      nblocks = all_blocks_len
    else
      ! ignore all_blocks and all_blocks_len in non-adaptive block mode
      nblocks = PRODUCT(domain%blnum(1:3))
    end if
    
    if(present(opt_blocks_changed)) blocks_changed = opt_blocks_changed

    if(.not. associated(domain%balance%blkweights)) then
      allocate(domain%balance%blkweights(1:domain%blnum(1),1:domain%blnum(2),1:domain%blnum(3)), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate domain%balance%blkweights')
    end if

    if(.not. associated(domain%balance%blkweights_ps)) then
      allocate(domain%balance%blkweights_ps(0:PRODUCT(domain%blnum(1:3))), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate domain%balance%blkweights_ps')
    end if
    blkweights_ps => domain%balance%blkweights_ps

    ! collect the block weights to domain%balance%blkweights
    call fd4_part_sfc_exch_weights(domain, err)
    FD4_L1_ASSERT(err/=0,'fd4_part_sfc_exch_weights failed')

    
    if(blocks_changed .or. sfctype /= domain%balance%last_sfctype ) then
    
      ! check if domain%balance%hidx and domain%balance%rnk are large enough, if not, deallocate
      if(associated(domain%balance%hidx)) then
        if(ubound(domain%balance%hidx,1)<nblocks) then
          deallocate(domain%balance%hidx)
          FD4_L3_ASSERT(.not.associated(domain%balance%rnk),'rnk not associated')
          deallocate(domain%balance%rnk)
        end if
      end if

      ! check if domain%balance%hidx and domain%balance%rnk are allocated
      if(.not. associated(domain%balance%hidx)) then
        allocate(domain%balance%hidx(nblocks))
        allocate(domain%balance%rnk(nblocks))
        blocks_changed = .true.
      end if

      ! get index of each block on the space-filling curve
      if(domain%varinfo%vthres_mode) then
        ! adaptive block mode: loop over specified blocks
        do ib=1,all_blocks_len
          call fd4_part_sfc_c2i(domain, sfctype, all_blocks(1:3,ib), domain%balance%hidx(ib))
        end do
      else
        ! non-adaptive block mode: loop over all blocks
        ib = 0
        do k=1,domain%blnum(3)
          pos(3) = k
          do j=1,domain%blnum(2)
            pos(2) = j
            do i=1,domain%blnum(1)
              pos(1) = i
              ib = ib + 1
              call fd4_part_sfc_c2i(domain, sfctype, pos, domain%balance%hidx(ib))
            end do
          end do
        end do
      end if

      ! rank the indices on the Hilbert curve, so that hidx(rnk(:)) is sorted
      call mrgrnk(domain%balance%hidx(1:nblocks), domain%balance%rnk(1:nblocks))
      
      domain%balance%last_sfctype = sfctype

    end if
    
    ! apply prefix sum to weight array
    maxweight = 0.0
    blkweights_ps(0) = 0.0
    do ib=1,nblocks
      if(domain%varinfo%vthres_mode) then
        ! all_blocks present
        pos(1:3) = all_blocks(1:3,domain%balance%rnk(ib))
      else
        ! compute block position from index in rnk
        pos(1:3) = fd4_domain_i2ccf(domain,domain%balance%rnk(ib))
      end if
      weight = domain%balance%blkweights( pos(1), pos(2), pos(3) )
      blkweights_ps(ib) = weight + blkweights_ps(ib-1)
      maxweight = MAX(maxweight, weight)
    end do

    if(method==FD4_PART_SFC_BINSRCH .or. method==FD4_PART_SFC_PARALLEL) then
      accuracy = 0.99_r8k
      binsteps = 100
      if(present(opt_accuracy)) accuracy=MAX(MIN(opt_accuracy,1.0_r8k),0.1_r8k)
      if(present(opt_binsteps)) binsteps=MAX(MIN(opt_binsteps,10000),1)
      ! compute the bottleneck error value which is necessary to get
      !  (accuracy * optimal_balance)
      ! as the resulting loadbalance of the 1d partitioning
      !  (eps is 0 if accuracy is 1.0: in this case the methods is exact binsteps is large enough)
      eps = (1.0_r8k-accuracy) * blkweights_ps(nblocks) / (accuracy*nprocs)
      ! this would be the necessary number of binary search steps for the given accuracy:
      !steps = log2i( CEILING( maxweight / eps, i_k ) )
      ! add statistics if necesary
      if(domain%balance%stat_binsteps<0) then
        call fd4_stat_add(domain%stat,'Balance max binsteps'  ,domain%balance%stat_binsteps)
      end if
    end if
    
    !! call partitioning routine

    if(method==FD4_PART_SFC_BINSRCH) then
      ! good serial method
      call fd4_part_1d_binsrch(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, &
                                       partition, maxload, reqsteps)
      domain%stat%values(domain%balance%stat_binsteps) = MAX(domain%stat%values(domain%balance%stat_binsteps), INT(reqsteps,i8k))
      !
    elseif(method==FD4_PART_SFC_PARALLEL) then    
      ! parallel method
      call fd4_part_1d_parallel(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, &
                                        domain%mpi, partition, maxload, reqsteps, err)
      domain%stat%values(domain%balance%stat_binsteps) = MAX(domain%stat%values(domain%balance%stat_binsteps), INT(reqsteps,i8k))
      !
    else
      ! simple method
      call fd4_part_1d_simple_h1(nblocks, nprocs, blkweights_ps(0:nblocks), partition, maxload)
      !
    end if

    balance = REAL( blkweights_ps(nblocks) / (nprocs * maxload), r4k)

    !! finally apply partition to FD4

    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now
    if(nblocks<PRODUCT(domain%blnum)) domain%blkmap(:,:,:,domain%blkmap_now) = -1
    ! set block map
    if(domain%varinfo%vthres_mode) then
      do crank=0,nprocs-1
        do ib=partition(crank),partition(crank+1)-1
          pos(1:3) = all_blocks(1:3,domain%balance%rnk(ib))
          domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
        end do
      end do
    else
      do crank=0,nprocs-1
        do ib=partition(crank),partition(crank+1)-1
          pos(1:3) = fd4_domain_i2ccf(domain,domain%balance%rnk(ib))
          domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
        end do
      end do
    end if
    ! set local list of blocks
    pstart = partition(rank)
    pend   = partition(rank+1)-1
    if(pstart<=pend) then
      domain%balance%sfc_pstart = pstart
      domain%balance%localweight = REAL(blkweights_ps(pend) - blkweights_ps(pstart-1),r4k)
      call stack3_create(local_blocks, pend - pstart + 1)
      local_blocks%pos = pend - pstart + 1
      if(domain%varinfo%vthres_mode) then
        do ib=1,local_blocks%pos
          local_blocks%a(1:3,ib) = all_blocks(1:3,domain%balance%rnk(ib+pstart-1))
        end do
      else
        do ib=1,local_blocks%pos
          local_blocks%a(1:3,ib) = fd4_domain_i2ccf(domain,domain%balance%rnk(ib+pstart-1))
        end do
      end if
    else
      domain%balance%sfc_pstart = -1
      domain%balance%localweight = 0.0
      local_blocks%pos = 0
    end if
    ! fill statistics
    domain%stat%sfcbnd(1) = MIN(domain%stat%sfcbnd(1), pstart)
    domain%stat%sfcbnd(2) = MAX(domain%stat%sfcbnd(2), pend)
    
    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    call fd4_part_sfc_blkmap_boundary(domain)    
    
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6,E13.4)') FD4_LINEHD,'new partition: ',local_blocks%pos, domain%balance%localweight
#endif
   
  
    err = 0

  end subroutine fd4_part_sfc_weight



  !! Exchange the weight of my blocks with all other processes.
  !!
  !! Updates domain%balance%blkweights.
  !!
  !! Internal routine.
  subroutine fd4_part_sfc_exch_weights(domain, err, opt_localweights, opt_set_blkweights)
    type(fd4_domain), intent(inout), target :: domain       !! the domain
    integer(i_k),     intent(out)           :: err          !! error status: 0...ok
    real(r4k),        intent(in),optional   :: opt_localweights(:) !! optional array of local block weights
    logical,          intent(in),optional   :: opt_set_blkweights  !! set $domain%$blkweights? default: yes, if allocated
    
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: crank, nblocks, k, j, i
    type(fd4_iter) :: iter
    logical :: set_blkweights

    mpi => domain%mpi
    
    mpi%recvcnt(:) = 0
    nblocks = 0
    err = -1
    
    set_blkweights = .true.
    if(present(opt_set_blkweights)) set_blkweights = opt_set_blkweights
    
    ! loop over blkmap to build recv counts for MPI_Allgatherv
    ! (number of blocks for each process)
    do k=1,domain%blnum(3)
      do j=1,domain%blnum(2)
        do i=1,domain%blnum(1)
          crank = domain%blkmap(i,j,k,domain%blkmap_now)
          if( crank >= 0 ) then
            mpi%recvcnt(crank) = mpi%recvcnt(crank) + 1
          end if
        end do
      end do
    end do

    FD4_L3_ASSERT(mpi%recvcnt(mpi%rank)/=domain%blocks%length,'nblocks not as expected')
    
    ! build the recv displacements
    mpi%recvdis(0) = 0
    do crank=1,mpi%nprocs-1
      mpi%recvdis(crank) = mpi%recvcnt(crank-1) + mpi%recvdis(crank-1)
    end do

    nblocks = mpi%recvdis(mpi%nprocs-1) + mpi%recvcnt(mpi%nprocs-1)

    ! this may fail, because domain%blockcount may be inconsistent
    ! after calling fd4_util_put_array
    !FD4_L3_ASSERT(nblocks/=domain%blockcount,'nblocks not as expected')
    
    ! allocate array for receiving the block weights
    if( size(mpi%recvreal,1) < nblocks ) then
      deallocate(mpi%recvreal)
      allocate(mpi%recvreal(nblocks))
    end if

    if(present(opt_localweights)) then
      !! use subroutine argument as local weights
      
      FD4_L3_ASSERT(size(opt_localweights)<mpi%recvcnt(mpi%rank),'too few weights')

      ! exchange the block weights
      call MPI_Allgatherv (opt_localweights(1), mpi%recvcnt(mpi%rank),                 MPI_R4K, &
                           mpi%recvreal(1),     mpi%recvcnt(0),        mpi%recvdis(0), MPI_R4K, mpi%comm, err )
      err = err - MPI_SUCCESS
      FD4_L1_ASSERT(err/=0,'MPI_Allgatherv failed')
      
    else
      !! fill mpi%sendreal with the local weights

      ! allocate array for sending the block weights
      if( size(mpi%sendreal,1) < mpi%recvcnt(mpi%rank) ) then
        deallocate(mpi%sendreal)
        allocate(mpi%sendreal(mpi%recvcnt(mpi%rank)))
      end if

      ! get the local block weights and write them to mpi%sendreal
      i = 0
      iter%cur => rbtree_fd4_block_iter_first(iter%biter, domain%blocks)
      do while(associated(iter%cur))
        i = i + 1
        mpi%sendreal(i) = iter%cur%weight
        iter%cur => rbtree_fd4_block_iter_next(iter%biter)
      end do
    
      FD4_L3_ASSERT(i/=domain%blocks%length,'nblocks not as expected')
      
      ! exchange the block weights
      call MPI_Allgatherv (mpi%sendreal(1), mpi%recvcnt(mpi%rank),                 MPI_R4K, &
                           mpi%recvreal(1), mpi%recvcnt(0),        mpi%recvdis(0), MPI_R4K, mpi%comm, err )
      err = err - MPI_SUCCESS
      FD4_L1_ASSERT(err/=0,'MPI_Allgatherv failed')
    
    end if
   
    
    if(set_blkweights .and. associated(domain%balance%blkweights)) then

      ! loop over blkmap and fill domain%balance%blkweights
      ! this method assumes the loop order in this loop has the same ordering
      ! as the index used to sort the blocks in domain%blocks
      do k=1,domain%blnum(3)
        do j=1,domain%blnum(2)
          do i=1,domain%blnum(1)
            crank = domain%blkmap(i,j,k,domain%blkmap_now)
            if( crank >= 0 ) then
              domain%balance%blkweights(i,j,k) = mpi%recvreal( mpi%recvdis(crank)+1 )
              mpi%recvdis(crank) = mpi%recvdis(crank) + 1
            else
              ! set default weight
              !TODO: do this only if the block has been removed last load balancing
              !      (or block will be added this time)
              domain%balance%blkweights(i,j,k) = domain%balance%defaultweight  &
               * REAL( PRODUCT( fd4_domain_blext_ijk(domain,i,j,k) ), r4k )
            end if
          end do
        end do
      end do

#if ( FD4_VERBOSE_LEVEL >= 3 )
      do crank=1,mpi%nprocs-1
        if(mpi%recvdis(crank) - mpi%recvdis(crank-1) /= mpi%recvcnt(crank)) then
          FD4_L1_ERROR('inconsistency after loop over blkmap')
        end if
      end do
#endif

    end if ! associated(domain%balance%blkweights)
    
    err = 0
    
  end subroutine fd4_part_sfc_exch_weights
  



  !! Create a new partition using block position. Weights are not considered.
  !!
  !! The special option $add_only effects:
  !! - all existing blocks (in blkmap) are left where they are
  !! - new blocks are added as if the option was not set
  !! - this (intentionally) results in a non balanced partitioning, but minimizes transfer costs
  !!   when - in a second partitioning step - blocks are migrated
  !!
  subroutine fd4_part_sfc_noweight(domain, all_blocks, all_blocks_len, local_blocks, sfctype, add_only, opt_blocks_changed)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: all_blocks_len                 !! length of $all_blocks(3,:)
    integer(i_k)    , intent(in)    :: all_blocks(:,:)                !! array of positions of the blocks in the new partion
    type(stack3)    , intent(inout) :: local_blocks                   !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    logical,optional, intent(in)    :: add_only                       !! only add *new* blocks, but don't touch existing blocks
    logical,optional, intent(in)    :: opt_blocks_changed             !! tell me if the blocks have changed (created/removed)
    
    integer :: rank, nprocs
    integer(i_k) :: i, crank, psum, pos(3)
    logical :: ao
    logical :: blocks_changed
    
    FD4_L1_ASSERT(all_blocks_len==0,'need allblocks')
    
    if(present(add_only)) then
      ao = add_only
    else
      ao = .false.
    end if
    
    rank   = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    blocks_changed = .true.
    
    if(present(opt_blocks_changed)) blocks_changed = opt_blocks_changed
    
    ! check if domain%balance%hidx and domain%balance%rnk are large enough, if not, deallocate
    if(associated(domain%balance%hidx)) then
      if(ubound(domain%balance%hidx,1)<all_blocks_len) then
        deallocate(domain%balance%hidx)
        FD4_L3_ASSERT(.not.associated(domain%balance%rnk),'rnk not associated')
        deallocate(domain%balance%rnk)
      end if
    end if

    ! check if domain%balance%hidx and domain%balance%rnk are allocated
    if(.not. associated(domain%balance%hidx)) then
      allocate(domain%balance%hidx(all_blocks_len))
      allocate(domain%balance%rnk(all_blocks_len))
      blocks_changed = .true.
    end if

    if(blocks_changed .or. sfctype /= domain%balance%last_sfctype ) then

      ! get index of each block on the space-filling curve
      do i=1,all_blocks_len
        call fd4_part_sfc_c2i(domain, sfctype, all_blocks(1:3,i), domain%balance%hidx(i))
      end do

      ! rank the indices on the hilbert curve, so that hidx(rnk(:)) is sorted
      call mrgrnk(domain%balance%hidx(1:all_blocks_len), domain%balance%rnk(1:all_blocks_len))

      domain%balance%last_sfctype = sfctype

    end if
    
    if(ao) domain%balance%last_sfctype = -1
        
    
    call stack3_create(local_blocks, all_blocks_len/nprocs)
    crank = 0
    psum = 0
    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now
    if(all_blocks_len<PRODUCT(domain%blnum)) domain%blkmap(:,:,:,domain%blkmap_now) = -1
    domain%balance%sfc_pstart = -1
    do i=1,all_blocks_len
      psum = psum + nprocs
      if(all_blocks_len*(crank+1)<psum) crank=crank+1
      pos(1:3) = all_blocks(1:3,domain%balance%rnk(i))
      if(ao .and. domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_last ) /= -1 ) then
        if(domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_last ) == domain%mpi%rank ) then
          call stack3_push(local_blocks, pos(1:3))
        end if
        domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_last )
      else
        domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
        if(rank == crank) then
          call stack3_push(local_blocks, pos(1:3))
          if(domain%balance%sfc_pstart<0) domain%balance%sfc_pstart = i
        end if
      end if
    end do
    
    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    call fd4_part_sfc_blkmap_boundary(domain)    
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6)') FD4_LINEHD,'new partition: ',local_blocks%pos
#endif
  
  end subroutine fd4_part_sfc_noweight


  !! Apply periodic boundary on blkmap.
  !!
  !! Internal routine.
  subroutine fd4_part_sfc_blkmap_boundary(domain)
    type(fd4_domain), intent(inout) :: domain  !! the domain

    integer(i_k) :: pos(3)

    pos = domain%blnum
    if(domain%periodic(1)) then
      domain%blkmap(0,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(pos(1),1:pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(pos(1)+1,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(1,1:pos(2),1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(2)) then
      domain%blkmap(1:pos(1),0,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),pos(2)+1,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),1,1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(3)) then
      domain%blkmap(1:pos(1),1:pos(2),0,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),1:pos(2),pos(3)+1,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),1,domain%blkmap_now)
    end if      

  end subroutine fd4_part_sfc_blkmap_boundary


  !! Get index of block at position $pos on the SFC of type $sfctype.
  !!
  !! The index is *not contiguous* over the exact size of the domain.
  !! i.e. $max($idx) == 3 ** (2**$domain%$sfclevel) (3D case)
  !!
  !! Internal routine.
  pure subroutine fd4_part_sfc_c2i(domain, sfctype, pos, idx)
    type(fd4_domain), intent(in)    :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    integer(i_k)    , intent(in)    :: pos(3)                         !! block position
    integer(i_k)    , intent(out)   :: idx                            !! index on SFC
  
    if(domain%balance%map2d(1) < 0) then
      ! 3D SFC
      if(sfctype==FD4_PART_SFC_MORTON) then
        call morton_c2i(domain%sfclevel,pos,idx)
      else
        call hilbert_c2i(domain%sfclevel,pos,idx)
      end if
    else
      ! 2D SFC
      if(sfctype==FD4_PART_SFC_MORTON) then
        call morton2d_c2i(domain%sfclevel,pos(domain%balance%map2d),idx)
      else
        call hilbert2d_c2i(domain%sfclevel,pos(domain%balance%map2d),idx)
      end if
    end if

  end subroutine fd4_part_sfc_c2i


  !! Get index of block at position $pos on the SFC of type $sfctype.
  !!
  !! The index is *contiguous* over the exact size of the domain.
  !! i.e. $max($idx) == $PRODUCT($domain%$blnum).
  !!
  !! Internal routine.
  pure subroutine fd4_part_sfc_c2i_qsz(domain, sfctype, pos, idx)
    type(fd4_domain), intent(in)    :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    integer(i_k)    , intent(in)    :: pos(3)                         !! block position
    integer(i_k)    , intent(out)   :: idx                            !! index on SFC
  
    if(domain%balance%map2d(1) < 0) then
      ! 3D SFC
      if(sfctype==FD4_PART_SFC_MORTON) then
        call morton_c2i_qsz(domain%sfclevel,domain%blnum,pos,idx)
      else
        call hilbert_c2i_qsz(domain%sfclevel,domain%blnum,pos,idx)
      end if
    else
      ! 2D SFC
      if(sfctype==FD4_PART_SFC_MORTON) then
        call morton2d_c2i_qsz(domain%sfclevel,domain%blnum(domain%balance%map2d),pos(domain%balance%map2d),idx)
      else
        call hilbert2d_c2i_qsz(domain%sfclevel,domain%blnum(domain%balance%map2d),pos(domain%balance%map2d),idx)
      end if
    end if

  end subroutine fd4_part_sfc_c2i_qsz


  !! Get position $pos of block at specified index $idx on the SFC of type $sfctype.
  !!
  !! The index is *contiguous* over the exact size of the domain,
  !! i.e. $max($idx) == $PRODUCT($domain%$blnum).
  !!
  !! Internal routine.
  pure subroutine fd4_part_sfc_i2c_qsz(domain, sfctype, idx, pos)
    type(fd4_domain), intent(in)    :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    integer(i_k)    , intent(in)    :: idx                            !! index on SFC
    integer(i_k)    , intent(out)   :: pos(3)                         !! block position
    
    integer(i_k) :: p(2)
  
    if(domain%balance%map2d(1) < 0) then
      ! 3D SFC
      if(sfctype==FD4_PART_SFC_MORTON) then
        call morton_i2c_qsz(domain%sfclevel,domain%blnum,idx,pos)
      else
        call hilbert_i2c_qsz(domain%sfclevel,domain%blnum,idx,pos)
      end if
    else
      ! 2D SFC
      if(sfctype==FD4_PART_SFC_MORTON) then
        call morton2d_i2c_qsz(domain%sfclevel,domain%blnum(domain%balance%map2d),idx,p)
      else
        call hilbert2d_i2c_qsz(domain%sfclevel,domain%blnum(domain%balance%map2d),idx,p)
      end if
      pos(1:3) = 1
      pos(domain%balance%map2d) = p
    end if

  end subroutine fd4_part_sfc_i2c_qsz
  


  !! Initialize SFC iterator with type sfctype.
  !!
  !! Internal routine.
  pure subroutine fd4_part_sfc_iter_init(iter, domain, sfctype)
    type(part_sfc_iter), intent(inout), target :: iter          !! SFC iterator
    type(fd4_domain)   , intent(in)            :: domain        !! the domain
    integer(i_k)       , intent(in)            :: sfctype       !! space-filling curve to use

    iter%idx =  -1
    iter%p   =  -1
    iter%sfctype = sfctype
    iter%map2d = domain%balance%map2d

    if(iter%map2d(1) < 0) then
      ! 3D SFC
      if(iter%sfctype==FD4_PART_SFC_MORTON) then
        call morton_iter_init(iter%miter, domain%sfclevel, domain%blnum)
        iter%p   = iter%miter%p
        iter%idx = iter%miter%idx
      else
        call hilbert_iter_init(iter%hiter, domain%sfclevel, domain%blnum)
        iter%p   = iter%hiter%p
        iter%idx = iter%hiter%idx
      end if
    else
      ! 2D SFC
      if(iter%sfctype==FD4_PART_SFC_MORTON) then
        call morton2d_iter_init(iter%m2diter, domain%sfclevel, domain%blnum(iter%map2d))
        iter%p   = 1
        iter%p(iter%map2d) = iter%m2diter%p
        iter%idx = iter%m2diter%idx
      else
        call hilbert2d_iter_init(iter%h2diter, domain%sfclevel, domain%blnum(iter%map2d))
        iter%p   = 1
        iter%p(iter%map2d) = iter%h2diter%p
        iter%idx = iter%h2diter%idx
      end if
    end if

  end subroutine fd4_part_sfc_iter_init


  !! Iterate to next position on SFC.
  !!
  !! Internal routine.
  pure subroutine fd4_part_sfc_iter_next(iter)
    type(part_sfc_iter), intent(inout), target :: iter          !! SFC iterator

    if(iter%map2d(1) < 0) then
      ! 3D SFC
      if(iter%sfctype==FD4_PART_SFC_MORTON) then
        call morton_iter_next(iter%miter)
        iter%p   = iter%miter%p
        iter%idx = iter%miter%idx
      else
        call hilbert_iter_next(iter%hiter)
        iter%p   = iter%hiter%p
        iter%idx = iter%hiter%idx
      end if
    else
      ! 2D SFC
      if(iter%sfctype==FD4_PART_SFC_MORTON) then
        call morton2d_iter_next(iter%m2diter)
        iter%p(iter%map2d) = iter%m2diter%p
        iter%idx = iter%m2diter%idx
      else
        call hilbert2d_iter_next(iter%h2diter)
        iter%p(iter%map2d) = iter%h2diter%p
        iter%idx = iter%h2diter%idx
      end if
    end if

  end subroutine fd4_part_sfc_iter_next

end module fd4_part_sfc_mod
