! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

! DO NOT CHANGE THESE ROUTINES WITHOUT THE SAME CHANGES IN stack.incf !

!! A simple stack of integers which grows ad infinutum if required.
!! Data is stored in a single contiguous array.
!!
!!> type(stack) :: st
!!  call stack_create(st,10)
!!  stack_push(st,4)
!!  stack_push(st,7)
!!  print *,'first element on stack is:',st%a(1)
!!< call stack_delete(st)
!!
!!
!!  Clear the stack, i.e. set topmost element = bottom.
!!  The size of the storage array is left unchanged.
!!v st%pos = 0
!!
!!
!!  Get element from stack.
!!  The index must be in the interval 1 <= $i <= size(st%a).
!!v st%a(i)
!!
!!  Add
!!v use stack_mod
!!  to your module/subroutine and
!!v include 'stack.incf'
!!  after the contains statement to enable inlining of the stack routines.
!!
!! @author Matthias Lieber
module stack_mod
  
  use kinds_mod
  implicit none
  
  private
  
  public :: stack
  
  type stack
    integer(i_k), pointer :: a(:) => null()  !! the actual data is stored in $a
    integer(i_k) :: pos = 0                  !! position of topmost element in the stack
  end type stack

contains

  !! Create a stack or reallocate at given size.
  !!
  !! If the stack is already allocated, the function reallocates
  !! the stack only in the case, that the current stack size is
  !! less then $inisize. Otherwise the stack is only cleared,
  !! i.e. $st%pos is set to 0.
  subroutine stack_create(st, inisize)
    type(stack), intent(inout) :: st         !! the stack
    integer(i_k), intent(in) :: inisize      !! initial size of array
    integer(i_k) :: is 
    is = MAX(inisize,1)
    if( associated(st%a) ) then
      if( size(st%a) < is ) then
        deallocate( st%a )
        allocate( st%a(is) )
      end if
    else
      allocate( st%a(is) )
    end if    
    st%pos = 0
  end subroutine stack_create

  !! Delete the stack and all of its memory.
  subroutine stack_delete(st)
    type(stack), intent(inout) :: st    !! the stack
    if(associated(st%a)) deallocate( st%a )
    st%a => null()
    st%pos = 0
  end subroutine stack_delete


  !! Reallocate the stack array to exactly fit the required size.
  subroutine stack_trim(st)
    type(stack), intent(inout) :: st    !! the stack
    integer(i_k), pointer :: a(:)
    if(.not. associated(st%a)) return
    if(st%pos<size(st%a)) then
      ! allocate smaller array
      allocate( a(MAX(1,st%pos)) )
      a(1:st%pos) = st%a(1:st%pos)
      deallocate(st%a)
      st%a => a
    end if
  end subroutine stack_trim  


  !! Push a new element to the top of the stack.
  !! The storage array will be enlarged by a factor of two if required.
  subroutine stack_push(st, v)
    type(stack), intent(inout) :: st    !! the stack
    integer(i_k), intent(in)  :: v      !! value to push to stack
    integer(i_k), pointer :: a(:)
    if(st%pos>=size(st%a)) then
      ! allocate bigger array
      allocate( a(2*st%pos) )
      a(1:st%pos) = st%a(1:st%pos)
      deallocate(st%a)
      st%a => a
      !print *,'allocated array of size',2*st%pos
    end if
    st%pos = st%pos + 1
    st%a( st%pos ) = v
  end subroutine stack_push  


  !! Push two new elements to the top of the stack.
  !! The storage array will be enlarged by a factor of two if required.
  subroutine stack_push2(st, v1, v2)
    type(stack), intent(inout) :: st    !! the stack
    integer(i_k), intent(in)  :: v1, v2 !! values to push to stack
    integer(i_k), pointer :: a(:)
    if(st%pos+1>=size(st%a)) then
      ! allocate bigger array
      allocate( a( max(2*st%pos,2+st%pos) ) )
      a(1:st%pos) = st%a(1:st%pos)
      deallocate(st%a)
      st%a => a
      !print *,'allocated array of size',2*st%pos
    end if
    st%pos = st%pos + 1
    st%a( st%pos ) = v1
    st%pos = st%pos + 1
    st%a( st%pos ) = v2
  end subroutine stack_push2 


  !! Push new elements to the top of the stack.
  !! The storage array will be enlarged by a factor of two if required.
  subroutine stack_push_array(st, v)
    type(stack), intent(inout) :: st       !! the stack
    integer(i_k), intent(in)  :: v(:)      !! values to push to stack
    integer(i_k), pointer :: a(:)
    integer(i_k) :: p
    p = size(v)
    if(st%pos+p>=size(st%a)) then
      ! allocate bigger array
      p = max(p, 2*st%pos)
      allocate( a(p) )
      a(1:st%pos) = st%a(1:st%pos)
      deallocate(st%a)
      st%a => a
      !print *,'allocated array of size',2*st%pos
    end if
    p = st%pos + 1
    st%pos = st%pos + size(v)
    st%a( p:st%pos ) = v(:)
  end subroutine stack_push_array


  !! Move (shallow copy) the data from $st_src to $st_dest.
  !! All data contained in $st_dest will be removed.
  !! $st_src will be empty on return.
  subroutine stack_move(st_src, st_dest)
    type(stack), intent(inout) :: st_src     !! the stack
    type(stack), intent(inout) :: st_dest    !! the stack
    call stack_delete(st_dest)
    ! shallow copy
    st_dest%a   => st_src%a
    st_dest%pos =  st_src%pos    
    ! empty st_src
    st_src%a   => null()
    st_src%pos = 0
  end subroutine stack_move


  !! Remove duplicates from $st, this is damn costly
  subroutine stack_remove_duplicates(st)
    type(stack), intent(inout) :: st     !! the stack
    integer :: i, j
    !! start of fancy O(N^2) algorithm
    i=1
    do while (i<st%pos)
      j=st%pos
      ! why starting at the end? because when swapping a hit with the last
      ! element, we are sure that the last element is no duplicate.
      do while (j>i)
        if( st%a(i) == st%a(j) ) then
          !! hit! swap with last element
          if(j<st%pos) st%a(j) = st%a(st%pos)
          st%pos = st%pos - 1
        end if
        j=j-1
      end do
      i=i+1
    end do    
  end subroutine stack_remove_duplicates


  !! Binary search!
  !! The first $n elements of the stack are searched for $v.
  !! Of course the stack must be sorted for the binary search to work!
  subroutine stack_binary_search(st, v, n, idx)
    type(stack),  intent(in)  :: st   !! the stack
    integer(i_k), intent(in)  :: v    !! value to find in the stack
    integer(i_k), intent(in)  :: n    !! number of elements to scan (upper bound)
    integer(i_k), intent(out) :: idx  !! position of found item, 0 if not found
    
    integer(i_k) :: m0, m1

    m0 = 0
    m1 = MIN(n,st%pos)+1

    do while ( m1-m0>1 )
      idx = ishft(m1-m0,-1) + m0   ! ishft instead of /2
      if(st%a(idx)<v) then
        m0 = idx
      else if(st%a(idx)>v) then
        m1 = idx
      else
        return
      end if
    end do
    
    idx = 0
    
  end subroutine stack_binary_search
  
end module stack_mod
